"""
Here we provide a representation for JSSP instances.

Our problem instances are actually extensions of
:class:`~numpy.ndarray`. They present the basic instance data as a
matrix, where each row corresponds to a job. Each row has twice the
number of machines elements. In an alternating fashion, we store the
machines that the job needs to go to as well as the time that it needs
on the machines, one by one.
Additionally, the instance name, the number of machines, and the number
of jobs are provided as attributes (although the latter two can easily
be inferred from the shape of the matrix).
Nevertheless, this memory layout and encapsulation as numpy array are
the most efficient way to store the data I could come up with.
"""
from importlib import resources  # nosem
from typing import Final

import numpy as np

import moptipy.utils.nputils as npu
from moptipy.api.component import Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import int_range_to_dtype
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import type_error

#: the recommended scope under which instance data should be stored
SCOPE_INSTANCE: Final = "inst"
#: The number of machines in the instance.
MACHINES: Final = "machines"
#: The number of jobs in the instance.
JOBS: Final = "jobs"
#: The lower bound of the makespan of the instance.
MAKESPAN_LOWER_BOUND: Final = "makespanLowerBound"
#: The upper bound of the makespan of the instance.
MAKESPAN_UPPER_BOUND: Final = "makespanUpperBound"

#: the internal final set of instances
_INSTANCES: Final[tuple[str, ...]] = \
    ("abz5", "abz6", "abz7", "abz8", "abz9",
     "demo",
     "dmu01", "dmu02", "dmu03", "dmu04", "dmu05", "dmu06", "dmu07",
     "dmu08", "dmu09", "dmu10", "dmu11", "dmu12", "dmu13", "dmu14",
     "dmu15", "dmu16", "dmu17", "dmu18", "dmu19", "dmu20", "dmu21",
     "dmu22", "dmu23", "dmu24", "dmu25", "dmu26", "dmu27", "dmu28",
     "dmu29", "dmu30", "dmu31", "dmu32", "dmu33", "dmu34", "dmu35",
     "dmu36", "dmu37", "dmu38", "dmu39", "dmu40", "dmu41", "dmu42",
     "dmu43", "dmu44", "dmu45", "dmu46", "dmu47", "dmu48", "dmu49",
     "dmu50", "dmu51", "dmu52", "dmu53", "dmu54", "dmu55", "dmu56",
     "dmu57", "dmu58", "dmu59", "dmu60", "dmu61", "dmu62", "dmu63",
     "dmu64", "dmu65", "dmu66", "dmu67", "dmu68", "dmu69", "dmu70",
     "dmu71", "dmu72", "dmu73", "dmu74", "dmu75", "dmu76", "dmu77",
     "dmu78", "dmu79", "dmu80",
     "ft06", "ft10", "ft20",
     "la01", "la02", "la03", "la04", "la05", "la06", "la07",
     "la08", "la09", "la10", "la11", "la12", "la13", "la14",
     "la15", "la16", "la17", "la18", "la19", "la20", "la21",
     "la22", "la23", "la24", "la25", "la26", "la27", "la28",
     "la29", "la30", "la31", "la32", "la33", "la34", "la35",
     "la36", "la37", "la38", "la39", "la40",
     "orb01", "orb02", "orb03", "orb04", "orb05", "orb06", "orb07",
     "orb08", "orb09", "orb10",
     "swv01", "swv02", "swv03", "swv04", "swv05", "swv06", "swv07",
     "swv08", "swv09", "swv10", "swv11", "swv12", "swv13", "swv14",
     "swv15", "swv16", "swv17", "swv18", "swv19", "swv20",
     "ta01", "ta02", "ta03", "ta04", "ta05", "ta06", "ta07",
     "ta08", "ta09", "ta10", "ta11", "ta12", "ta13", "ta14",
     "ta15", "ta16", "ta17", "ta18", "ta19", "ta20", "ta21",
     "ta22", "ta23", "ta24", "ta25", "ta26", "ta27", "ta28",
     "ta29", "ta30", "ta31", "ta32", "ta33", "ta34", "ta35",
     "ta36", "ta37", "ta38", "ta39", "ta40", "ta41", "ta42",
     "ta43", "ta44", "ta45", "ta46", "ta47", "ta48", "ta49",
     "ta50", "ta51", "ta52", "ta53", "ta54", "ta55", "ta56",
     "ta57", "ta58", "ta59", "ta60", "ta61", "ta62", "ta63",
     "ta64", "ta65", "ta66", "ta67", "ta68", "ta69", "ta70",
     "ta71", "ta72", "ta73", "ta74", "ta75", "ta76", "ta77",
     "ta78", "ta79", "ta80",
     "yn1", "yn2", "yn3", "yn4")


# start lb
def compute_makespan_lower_bound(machines: int,
                                 jobs: int,
                                 matrix: np.ndarray) -> int:
    """
    Compute the lower bound for the makespan of a JSSP instance data.

    :param machines: the number of machines
    :param jobs: the number of jobs
    :param matrix: the matrix with the instance data
    :returns: the lower bound for the makespan
    """
    # get the lower bound of the makespan with the algorithm by Taillard
    usedmachines = np.zeros(machines, np.bool_)  # -lb
    jobtimes = np.zeros(jobs, npu.DEFAULT_INT)  # get array for job times
    machinetimes = np.zeros(machines, npu.DEFAULT_INT)  # machine times array
    machine_start_idle = npu.np_ints_max(machines, npu.DEFAULT_INT)
    machine_end_idle = npu.np_ints_max(machines, npu.DEFAULT_INT)

    for jobidx in range(jobs):  # iterate over all jobs
        usedmachines.fill(False)  # no machine has been used  # -lb
        jobtime: int = 0  # the job time sum
        for i in range(machines):  # iterate over all operations
            machine, time = matrix[jobidx, i]  # get operation data
            if usedmachines[i]:  # machine already used??? -> error  # -lb
                raise ValueError(  # -lb
                    f"Machine {machine} occurs more than once.")  # -lb
            usedmachines[i] = True  # mark machine as used  # -lb
            if time < 0:  # time can _never_ be negative -> error  # -lb
                raise ValueError(f"Invalid time '{time}'.")  # -lb
            machinetimes[machine] += time  # add up operation times
            machine_start_idle[machine] = min(  # update with...
                machine_start_idle[machine], jobtime)  # ...job time
            jobtime += time  # update job time by adding operation time

        jobtimes[jobidx] = jobtime  # store job time
        jobremaining = jobtime  # iterate backwards to get end idle times
        for i in range(machines - 1, -1, -1):  # second iteration round
            machine, time = matrix[jobidx, i]  # get machine for operation
            machine_end_idle[machine] = min(  # update by computing...
                machine_end_idle[machine],  # the time that the job...
                jobtime - jobremaining)  # needs _after_ operation
            jobremaining -= time  # and update the remaining job time

        if not all(usedmachines):  # all machines have been used?  # -lb
            raise ValueError("Some machines not used in a job.")  # -lb

    # get the maximum of the per-machine sums of the idle and work times
    machines_bound = (machine_start_idle + machine_end_idle
                      + machinetimes).max()
    if machines_bound <= 0:  # -lb
        raise ValueError("Computed machine bound cannot be <= , "  # -lb
                         f"but is {machines_bound}.")  # -lb
    # get the longest time any job needs in total
    jobs_bound = jobtimes.max()
    if jobs_bound <= 0:  # -lb
        raise ValueError(  # -lb
            f"Computed jobs bound cannot be <= , but is {jobs_bound}.")  # -lb

    return int(max(machines_bound, jobs_bound))  # return bigger one
# end lb


# start book
class Instance(Component, np.ndarray):
    """
    An instance of the Job Shop Scheduling Problem.

    Besides the metadata, this object is a three-dimensional np.ndarray
    where the columns stand for jobs and the rows represent the
    operations of the jobs. Each row*column contains two values (third
    dimension), namely the machine where the operation goes and the time
    it will consume at that machine: `I[job, operation, 0] = machine`,
    `I[job, operation, 1] = time` that the job spents on machine.
    """

    # the name of the instance
    name: str
    #: the number of jobs == self.shape[0]
    jobs: int
    #: the number of machines == self.shape[1]
    machines: int
    # ... some more properties and methods ...
    # end book
    #: the lower bound of the makespan of this JSSP instance
    makespan_lower_bound: int
    #: the upper bound of the makespan of this JSSP instance
    makespan_upper_bound: int

    def __new__(cls, name: str, machines: int, jobs: int,
                matrix: np.ndarray,
                makespan_lower_bound: int | None = None) -> "Instance":
        """
        Create an instance of the Job Shop Scheduling Problem.

        :param cls: the class
        :param name: the name of the instance
        :param machines: the number of machines
        :param jobs: the number of jobs
        :param matrix: the matrix with the data (will be copied)
        :param makespan_lower_bound: the lower bound of the makespan, which
            may be the known global optimum if the instance has been solved
            to optimality or any other approximation. If `None` is provided,
            a lower bound will be computed.
        """
        use_name: Final[str] = sanitize_name(name)
        if name != use_name:
            raise ValueError(f"Name '{name}' is not a valid name.")

        if (not isinstance(machines, int)) or (machines < 1):
            raise ValueError("There must be at least one machine, "
                             f"but '{machines}' were specified in "
                             f"instance '{name}'.")
        if not isinstance(jobs, int) or (jobs < 1):
            raise ValueError("There must be at least one job, "
                             f"but '{jobs}' were specified in "
                             f"instance '{name}'.")
        if not isinstance(matrix, np.ndarray):
            raise type_error(matrix, "matrix", np.ndarray)

        use_shape: tuple[int, int, int] = (jobs, machines, 2)
        if matrix.shape[0] != jobs:
            raise ValueError(
                f"Invalid shape '{matrix.shape}' of matrix: must have "
                f"jobs={jobs} rows, but has {matrix.shape[0]} in "
                f"instance '{name}'.")
        if len(matrix.shape) == 3:
            if matrix.shape[1] != machines:
                raise ValueError(
                    f"Invalid shape '{matrix.shape}' of matrix: must have "
                    f"2*machines={machines} columns, but has "
                    f"{matrix.shape[1]} in instance '{name}'.")
            if matrix.shape[2] != 2:
                raise ValueError(
                    f"Invalid shape '{matrix.shape}' of matrix: must have "
                    f"2 cells per row/column tuple, but has "
                    f"{matrix.shape[2]} in instance '{name}'.")
        elif len(matrix.shape) == 2:
            if matrix.shape[1] != 2 * machines:
                raise ValueError(
                    f"Invalid shape '{matrix.shape}' of matrix: must have "
                    f"2*machines={2 * machines} columns, but has "
                    f"{matrix.shape[1]} in instance '{name}'.")
            matrix = matrix.reshape(use_shape)
        else:
            raise ValueError(
                "JSSP instance data matrix must have two or three"
                f"dimensions, but has {len(matrix.shape)} in instance "
                f"'{name}'.")
        if matrix.shape != use_shape:
            raise ValueError(
                f"matrix.shape is {matrix.shape}, not {use_shape}?")
        if not npu.is_np_int(matrix.dtype):
            raise ValueError(
                "Matrix must have an integer type, but is of type "
                f"'{matrix.dtype}' in instance '{name}'.")
        # ... some computations ...
        ms_lower_bound = compute_makespan_lower_bound(machines, jobs, matrix)
        ms_upper_bound = int(matrix[:, :, 1].sum())  # sum of all job times
        if ms_upper_bound < ms_lower_bound:
            raise ValueError(
                f"Computed makespan upper bound {ms_upper_bound} must not "
                f"be less than computed lower bound {ms_lower_bound}.")
        if makespan_lower_bound is None:
            makespan_lower_bound = ms_lower_bound
        else:
            if not isinstance(makespan_lower_bound, int):
                raise type_error(makespan_lower_bound,
                                 f"makespan lower bound of instance {name}",
                                 int)
            if makespan_lower_bound <= 0:
                raise ValueError("If specified, makespan_lower_bound must be "
                                 f"positive, but is {makespan_lower_bound} "
                                 f"in instance '{name}'.")
            if makespan_lower_bound < ms_lower_bound:
                raise ValueError(
                    "If specified, makespan_lower_bound must be >= "
                    f"{ms_lower_bound}, but is {makespan_lower_bound} in "
                    f"instance '{name}'.")
            if makespan_lower_bound > ms_upper_bound:
                raise ValueError(
                    "If specified, makespan_lower_bound must be <= "
                    f"{ms_upper_bound}, but is {makespan_lower_bound} in "
                    f"instance '{name}'.")
        obj: Final[Instance] = super().__new__(
            Instance, use_shape, int_range_to_dtype(
                min_value=0, max_value=int(matrix.max())))
        np.copyto(obj, matrix, casting="safe")
        #: the name of the instance
        obj.name = use_name
        #: the number of jobs == self.shape[0]
        obj.jobs = jobs
        #: the number of machines == self.shape[1]
        obj.machines = machines
        #: the lower bound of the makespan of this JSSP instance
        obj.makespan_lower_bound = makespan_lower_bound
        #: the upper bound of the makespan of this JSSP instance
        obj.makespan_upper_bound = ms_upper_bound
        return obj

    def __str__(self) -> str:
        """
        Get the name of this JSSP instance.

        :return: the name
        """
        return self.name

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters describing this JSSP instance to the logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value(MACHINES, self.machines)
        logger.key_value(JOBS, self.jobs)
        logger.key_value(MAKESPAN_LOWER_BOUND, self.makespan_lower_bound)
        logger.key_value(MAKESPAN_UPPER_BOUND, self.makespan_upper_bound)
        logger.key_value(npu.KEY_NUMPY_TYPE, self.dtype.char)

    @staticmethod
    def from_text(name: str, rows: list[str]) -> "Instance":
        """
        Convert a name and a set of rows of text to an JSSP instance.

        :param name: the name of the instance
        :param rows: the rows
        :return: the JSSP Instance
        """
        if not isinstance(rows, list):
            raise type_error(rows, "rows", list)
        if len(rows) < 3:
            raise ValueError(
                f"Must have at least 3 rows, but found {rows}.")
        description = rows[0]
        if not isinstance(description, str):
            raise type_error(description, "first element of rows", str)
        jobs_machines_txt = rows[1]

        matrix = np.asanyarray([np.fromstring(row, dtype=npu.DEFAULT_INT,
                                              sep=" ")
                                for row in rows[2:]])
        if not np.issubdtype(matrix.dtype, np.integer):
            raise ValueError("Error when converting array to matrix, "
                             f"got type '{matrix.dtype}'.")

        min_value = int(matrix.min())
        if min_value < 0:
            raise ValueError("JSSP matrix can only contain values >= 0, "
                             f"but found {min_value}.")

        max_value = int(matrix.max())
        if max_value <= min_value:
            raise ValueError(
                "JSSP matrix must contain value larger than minimum "
                f"{min_value}, but maximum is {max_value}.")

        dtype = int_range_to_dtype(min_value=min_value, max_value=max_value)
        if dtype != matrix.dtype:
            matrix = matrix.astype(dtype)

        jobs_machines = jobs_machines_txt.strip().split(" ")
        jobs = int(jobs_machines[0])
        machines = int(jobs_machines[len(jobs_machines) - 1])

        makespan_lower_bound = None
        i = description.find("lower bound:")
        if i >= 0:
            i += 12
            j = description.find(";", i)
            if j < i:
                j = len(description)
            makespan_lower_bound = int(description[i:j].strip())

        return Instance(name=name,
                        jobs=jobs,
                        machines=machines,
                        matrix=matrix,
                        makespan_lower_bound=makespan_lower_bound)

    @staticmethod
    def from_stream(name: str, stream) -> "Instance":
        """
        Load an instance from a text stream.

        :param name: the name of the instance to be loaded
        :param stream: the text stream
        :return: the instance
        """
        state = 0
        rows: list[str] | None = None
        for line in stream:
            line = str(line).strip()
            if len(line) <= 0:
                continue
            if state == 0:
                if line.startswith("+++"):
                    state = 1
                continue
            if state == 1:
                if line.startswith("instance "):
                    inst = line[9:].strip()
                    if inst == name:
                        state = 2
                    else:
                        state = 4
            elif state == 2:
                if line.startswith("+++"):
                    state = 3
                    rows = []
                    continue
                raise ValueError(f"Unexpected string '{line}'.")
            elif state == 3:
                if line.startswith("+++"):
                    return Instance.from_text(name=name,
                                              rows=rows)
                rows.append(line)
            elif state == 4:
                if line.startswith("+++"):
                    state = 5
                    continue
                raise ValueError(f"Unexpected string '{line}'.")
            elif state == 5:
                if line.startswith("+++"):
                    state = 1

        raise ValueError(f"Could not find instance '{name}'.")

    @staticmethod
    def from_resource(name: str) -> "Instance":
        """
        Load the JSSP instances `name` provided as part of moptipy.

        :param name: the instance name
        :return: the instance

        >>> jssp = Instance.from_resource("demo")
        >>> print(jssp.jobs)
        4
        >>> print(jssp.machines)
        5
        >>> print(jssp)
        demo
        """
        with resources.open_text(package=str(__package__),
                                 resource="demo.txt" if (name == "demo")
                                 else "instances.txt") as stream:
            return Instance.from_stream(name=name, stream=stream)

    @staticmethod
    def list_resources() -> tuple[str, ...]:
        """
        Get a tuple with all JSSP instances provided in the moptipy resources.

        :return: a tuple with all instance names that are valid parameters
            to :meth:`Instance.from_resource`

        >>> print(Instance.list_resources()[0:3])
        ('abz5', 'abz6', 'abz7')
        """
        return _INSTANCES


def check_instance(inst: Instance) -> Instance:
    """
    Check whether the contents of a JSSP instance are OK.

    This method thoroughly checks the contents of an instance and the
    types of all of its members. If your instances passes this method
    without any error, it is a valid JSSP instance that can be used for
    experiments. All instances in our benchmark set listed above will
    pass this test.

    :param inst: the instance
    :returns: the instance, if its contents are OK
    """
    if not isinstance(inst, Instance):
        raise type_error(inst, "instance", Instance)
    if not isinstance(inst, np.ndarray):
        raise type_error(inst, "instance", np.ndarray)
    if not isinstance(inst.machines, int):
        raise type_error(inst.machines, "inst.machines", int)
    if not isinstance(inst.name, str):
        raise type_error(inst.name, "inst.name", str)
    if (len(inst.name) <= 0) \
            or (inst.name != sanitize_name(inst.name)):
        raise ValueError(f"invalid instance name '{inst.name}'")
    if inst.machines <= 0:
        raise ValueError(f"inst.machines must be > 0, but is {inst.machines}"
                         f" for instance {inst.name}")
    if not isinstance(inst.jobs, int):
        raise type_error(inst.jobs, "inst.jobs", int)
    if inst.machines <= 0:
        raise ValueError(f"inst.jobs must be > 0, but is {inst.jobs}"
                         f" for instance {inst.name}.")
    if not isinstance(inst.makespan_lower_bound, int):
        raise type_error(inst.makespan_lower_bound,
                         f"inst.makespan_lower_bound of {inst.name}", int)
    if inst.makespan_lower_bound <= 0:
        raise ValueError("inst.makespan_lower_bound must be > 0,"
                         f" but is {inst.makespan_lower_bound}"
                         f" for instance {inst.name}.")
    if not isinstance(inst.makespan_upper_bound, int):
        raise type_error(inst.makespan_upper_bound,
                         f"inst.makespan_upper_bound of {inst.name}", int)
    if inst.makespan_upper_bound <= 0:
        raise ValueError("inst.makespan_upper_bound must be > 0,"
                         f" but is {inst.makespan_upper_bound}"
                         f" for instance {inst.name}.")
    if len(inst.shape) != 3:
        raise ValueError(f"inst must be 3d-array, but has shape {inst.shape}"
                         f" for instance {inst.name}.")
    if inst.shape[0] != inst.jobs:
        raise ValueError(
            f"inst.shape[0] must be inst.jobs={inst.jobs}, "
            f"but inst has shape {inst.shape} for instance {inst.name}.")
    if inst.shape[1] != inst.machines:
        raise ValueError(
            f"inst.machines[1] must be inst.machines={inst.machines}, "
            f"but inst has shape {inst.shape} for instance {inst.name}.")
    if inst.shape[2] != 2:
        raise ValueError(
            f"inst.machines[2] must be 2, but inst has shape {inst.shape}"
            f" for instance {inst.name}.")

    for i in range(inst.jobs):
        for j in range(inst.machines):
            machine = inst[i, j, 0]
            if not (0 <= machine < inst.machines):
                raise ValueError(
                    f"encountered machine {machine} for job {i} in "
                    f"operation {j}, but there are only "
                    f"{inst.machines} machines for instance {inst.name}.")
    mslb = compute_makespan_lower_bound(
        machines=inst.machines, jobs=inst.jobs, matrix=inst)
    if mslb > inst.makespan_lower_bound:
        raise ValueError(f"makespan lower bound computed as {mslb},"
                         f"but set to {inst.makespan_upper_bound},"
                         "which is not lower  for instance {inst.name}.")
    msub = sum(inst[:, :, 1].flatten())
    if msub != inst.makespan_upper_bound:
        raise ValueError(f"makespan upper bound computed as {msub}, "
                         f"but set to {inst.makespan_upper_bound}"
                         f" for instance {inst.name}.")
    tub = max([sum(inst[i, :, 1]) for i in range(inst.jobs)])
    if inst.makespan_lower_bound < tub:
        raise ValueError(f"makespan lower bound {inst.makespan_lower_bound} "
                         f"less then longest job duration {tub}"
                         f" for instance {inst.name}.")
    return inst



"""
The base classes for multi-objective optimization problems.

This class provides the ability to evaluate solutions according to multiple
criteria. The evaluation results are stored in a numpy array and also are
scalarized to a single value.

Basically, a multi-objective problem provides three essential components:

1. It can evaluate a candidate solution according to multiple optimization
   objectives. Each objective returns one value, subject to minimization,
   and all the values are stored in a single numpy array.
   This is done by :meth:`~moptipy.api.mo_problem.MOProblem.f_evaluate`
2. It provides a criterion deciding whether one such objective vector
   dominates (i.e., is strictly better than) another one. This is done by
   :meth:`~moptipy.api.mo_problem.MOProblem.f_dominates`. The default
   definition adheres to the standard "domination" definition in
   multi-objective optimization: A vector `a` dominates a vector `b` if it
   is not worse in any objective value and better in at least one. But if
   need be, you can overwrite this behavior.
3. A scalarization approach: When evaluating a solution, the result is not
   just the objective vector itself, but also a single scalar value. This is
   needed to create compatibility to single-objective optimization. Matter of
   fact, a :class:`~moptipy.api.mo_problem.MOProblem` is actually a subclass
   of :class:`~moptipy.api.objective.Objective`. This means that via this
   scalarization, all multi-objective problems can also be considered as
   single-objective problems. This means that single-objective algorithms can
   be applied to them as-is. It also means that log files are compatible.
   Multi-objective algorithms can just ignore the scalarization result and
   focus on the domination relationship. Often, a weighted sum approach
   (:class:`~moptipy.mo.problem.weighted_sum.WeightedSum`) may be the method
   of choice for scalarization.
"""
from typing import Final

import numpy as np

from moptipy.api.logging import KEY_SPACE_NUM_VARS, SCOPE_OBJECTIVE_FUNCTION
from moptipy.api.mo_utils import dominates
from moptipy.api.objective import Objective, check_objective
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import (
    DEFAULT_FLOAT,
    DEFAULT_INT,
    DEFAULT_UNSIGNED_INT,
    KEY_NUMPY_TYPE,
    int_range_to_dtype,
    val_numpy_type,
)
from moptipy.utils.types import type_error


class MOProblem(Objective):
    """
    The base class for multi-objective optimization problems.

    A multi-objective optimization problem is defined as a set of
    :class:`~moptipy.api.objective.Objective` functions. Each candidate
    solution is evaluated using each of the objectives, i.e., is rated by a
    vector of objective values. This vector is the basis for deciding which
    candidate solutions to keep and which to discard.

    In multi-objective optimization, this decision is based on "domination."
    A solution `a` dominates a solution `b` if it is not worse in any
    objective and better in at least one. This comparison behavior is
    implemented in method
    :meth:`~moptipy.api.mo_problem.MOProblem.f_dominates` and can be
    overwritten if need be.

    In our implementation, we prescribe that each multi-objective optimization
    problem must also be accompanied by a scalarization function, i.e., a
    function that represents the vector of objective values as a single scalar
    value. The whole multi-objective problem can then be viewed also as a
    single objective function itself. The method
    :meth:`~moptipy.api.mo_problem.MOProblem.evaluate` first evaluates all of
    the objective functions and obtains the vector of objective values. It then
    scalarizes the result into a single scalar quality and returns it.
    Multi-objective algorithms may instead use the method
    :meth:`~moptipy.api.mo_problem.MOProblem.f_evaluate`, which also allows
    a vector to be passed in which will then be filled with the results of the
    individual objective functions.

    This makes multi-objective optimization with moptipy compatible with
    single-objective optimization. In other words, all optimization methods
    implemented for single-objective processes
    :class:`~moptipy.api.process.Process` will work out-of-the-box with the
    multi-objective version :class:`~moptipy.api.mo_process.MOProcess`.

    Warning: We use instances of :class:`numpy.ndarray` to represent the
    vectors of objective values. This necessitates that each objective
    function has, if it is integer-valued
    (:meth:`~moptipy.api.objective.Objective.is_always_integer` is `True`)
    a range that fits well into at least a 64-bit integer. Specifically, it
    must be possible to compute "a - b" without overflow or loss of sign for
    any two objective values "a" and "b" within the confines of a numpy
    signed 64-bit integer.
    """

    def f_create(self) -> np.ndarray:
        """
        Create a vector to receive the objective values.

        This array will be of the length returned by :meth:`f_dimension` and
        of the `dtype` of :meth:`f_dtype`.

        :returns: a vector to receive the objective values
        """
        return np.empty(self.f_dimension(), self.f_dtype())

    def f_dtype(self) -> np.dtype:
        """
        Get the data type used in :meth:`f_create`.

        This data type will be an integer data type if all the objective
        functions are integer-valued. If the bounds of the objective values
        are known, then this type will be "big enough" to allow the
        subtraction "a - b" of any two objective vectors "a" and "b" to be
        computed without overflow or loss of sign. At most, however, this
        data type will be a 64-bit integer.
        If any one of the objective functions returns floating point data,
        this data type will be a floating point type.

        :returns: the data type used by :meth:`f_create`.
        """

    def f_dimension(self) -> int:
        """
        Obtain the number of objective functions.

        :returns: the number of objective functions
        """

    def f_validate(self, x: np.ndarray) -> None:
        """
        Validate the objective vector.

        :param x: the numpy vector
        :raises TypeError: if the string is not an element of this space.
        :raises ValueError: if the shape of the vector is wrong or any of its
            element is not finite.
        """
        if not isinstance(x, np.ndarray):
            raise TypeError(x, "x", np.ndarray)
        shape = x.shape
        if len(shape) != 1:
            raise ValueError(
                f"{x} cannot have more than one dimension, but has {shape}!")
        dim = self.f_dimension()  # pylint: disable=E1111
        if shape[0] != dim:
            raise ValueError(
                f"{x} should have length {dim} but has {shape[0]}!")
        dt = self.f_dtype()  # pylint: disable=E1111
        if x.dtype != dt:
            raise ValueError(f"{x} should have dtype {dt} but has {x.dtype}!")

    def f_evaluate(self, x, fs: np.ndarray) -> int | float:
        """
        Perform the multi-objective evaluation of a solution.

        This method fills the objective vector `fs` with the results of the
        objective functions evaluated on `x`. It then returns the scalarized
        result, i.e., a single scalar value computed based on all values
        in `fs`.

        :param x: the solution to be evaluated
        :param fs: the array to receive the objective values
        :returns: the scalarization result
        """

    # noinspection PyMethodMayBeStatic
    def f_dominates(self, a: np.ndarray, b: np.ndarray) -> int:
        """
        Check if an objective vector dominates or is dominated by another one.

        Usually, one vector is said to dominate another one if it is not worse
        in any objective and better in at least one. This behavior is
        implemented in :func:`moptipy.api.mo_utils.dominates` and this is also
        the default behavior of this method. However, depending on your
        concrete optimization task, you may overwrite this behavior.

        :param a: the first objective vector
        :param b: the second objective value
        :returns: an integer value indicating the domination relationship
        :retval -1: if `a` dominates `b`
        :retval 1: if `b` dominates `a`
        :retval 2: if `b` equals `a`
        :retval 0: if `a` and `b` are mutually non-dominated, i.e., if neither
            `a` dominates `b` not `b` dominates `a` and `b` is also different
            from `a`
        """
        return dominates(a, b)

    def evaluate(self, x) -> float | int:
        """
        Evaluate a solution `x` and return its scalarized objective value.

        This method computes all objective values for a given solution and
        then returns the scalarized result. The objective values themselves
        are directly discarted and not used. It makes a multi-objective
        problem compatible with single-objective optimization.

        :param x: the candidate solution
        :returns: the scalarized objective value
        """
        return self.f_evaluate(x, self.f_create())

    def __str__(self) -> str:
        """Get the string representation of this scalarization."""
        return "moProblem"


def check_mo_problem(mo_problem: MOProblem) -> MOProblem:
    """
    Check whether an object is a valid instance of :class:`MOProblem`.

    :param mo_problem: the multi-objective optimization problem
    :return: the mo-problem
    :raises TypeError: if `mo_problem` is not an instance of
        :class:`MOProblem`
    """
    check_objective(mo_problem)
    if not isinstance(mo_problem, MOProblem):
        raise type_error(mo_problem,
                         "multi-objective optimziation problem", MOProblem)
    return mo_problem


class MOSOProblemBridge(MOProblem):
    """A bridge between multi-objective and single-objective optimization."""

    def __init__(self, objective: Objective) -> None:
        """Initialize the bridge."""
        super().__init__()
        check_objective(objective)

        self.evaluate = objective.evaluate  # type: ignore
        self.lower_bound = objective.lower_bound  # type: ignore
        self.upper_bound = objective.upper_bound  # type: ignore
        self.is_always_integer = objective.is_always_integer  # type: ignore

        dt: np.dtype
        if self.is_always_integer():
            lb: int | float = self.lower_bound()
            ub: int | float = self.upper_bound()
            dt = DEFAULT_INT
            if isinstance(lb, int):
                if isinstance(ub, int):
                    dt = int_range_to_dtype(lb, ub)
                elif lb >= 0:
                    dt = DEFAULT_UNSIGNED_INT
        else:
            dt = DEFAULT_FLOAT

        #: the data type of the objective array
        self.__dtype: Final[np.dtype] = dt
        #: the objective function
        self.__f: Final[Objective] = objective
        self.f_create = lambda dd=dt: np.empty(1, dd)  # type: ignore
        self.f_dimension = lambda: 1  # type: ignore

    def f_evaluate(self, x, fs: np.ndarray) -> int | float:
        """
        Evaluate the candidate solution.

        :param x: the solution
        :param fs: the objective vector, will become `[res]`
        :returns: the objective value `res`
        """
        res: Final[int | float] = self.evaluate(x)
        fs[0] = res
        return res

    def f_dtype(self) -> np.dtype:
        """Get the objective vector dtype."""
        return self.__dtype

    def f_validate(self, x: np.ndarray) -> None:
        """
        Validate the objective vector.

        :param x: the numpy array with the objective values
        """
        if not isinstance(x, np.ndarray):
            raise type_error(x, "x", np.ndarray)
        if len(x) != 1:
            raise ValueError(f"length of x={len(x)}")
        lb = self.lower_bound()
        ub = self.upper_bound()
        if not (lb <= x[0] <= ub):
            raise ValueError(f"failed: {lb} <= {x[0]} <= {ub}")

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this function to the provided destination.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value(KEY_SPACE_NUM_VARS, "1")
        logger.key_value(KEY_NUMPY_TYPE, val_numpy_type(self.__dtype))
        with logger.scope(f"{SCOPE_OBJECTIVE_FUNCTION}{0}") as scope:
            self.__f.log_parameters_to(scope)



"""An implementation of a search space for permutations of a base string."""
from math import factorial
from typing import Final, Iterable

import numpy

from moptipy.spaces.intspace import IntSpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import array_to_str
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import type_error

#: the base string to be permuted
KEY_BASE_STRING: Final[str] = "baseString"
#: the number of times each value must occur
KEY_REPETITIONS: Final[str] = "repetitions"


class Permutations(IntSpace):  # +book
    """
    A space of permutations of a base string.

    This class includes standard permutations of the form 0, 1, 2, ..., n-1,
    but also permutations with repetitions.
    The idea is that a base string is defined, i.e., an array of integer
    values. In this array, some values may appear twice, some may be missing.
    For example, `[1, 3, 5, 5, 7]` is a proper base string. The space defined
    over this base string would then contain values such as `[1, 3, 5, 5, 7]`,
    `[1, 5, 3, 5, 7]`, `[7, 5, 5, 3, 1]` and so on. Basically, it will contain
    all strings that can be created by shuffling the base string. These
    strings have in common that they contain exactly all the values from the
    base string and contain them exactly as same as often as they appear in
    the base string. The space defined upon the above base string therefore
    would contain 5! / (1! * 1! * 2! * 1!) = 120 / 2 = 60 different strings.

    The permutation space defined above can be created as follows:

    >>> perm = Permutations([1, 3, 5, 5, 7])
    >>> print(perm.to_str(perm.blueprint))
    1;3;5;5;7
    >>> print(perm)
    perm_of_string
    >>> print(perm.n_points())
    60

    Another example is this:

    >>> perm = Permutations((1, 2, 3, 3, 2))
    >>> print(perm.to_str(perm.blueprint))
    1;2;2;3;3
    >>> print(perm)
    perm_of_string
    >>> print(perm.n_points())
    30

    If you want to create a permutation with repetitions, e.g.,
    where each of the n=4 values from 0 to 3 appear exactly 3 times,
    you can use the utility method `with_repetitions`:

    >>> perm = Permutations.with_repetitions(4, 3)
    >>> print(perm.to_str(perm.blueprint))
    0;0;0;1;1;1;2;2;2;3;3;3
    >>> print(perm)
    perm4w3r
    >>> print(perm.n_points())
    369600

    If you instead want to create standard permutations, i.e., where
    each value from 0 to n-1 appears exactly once, you would do:

    >>> perm = Permutations.standard(5)
    >>> print(perm.to_str(perm.blueprint))
    0;1;2;3;4
    >>> print(perm)
    perm5
    >>> print(perm.n_points())
    120
    """

    def __init__(self, base_string: Iterable[int]) -> None:  # +book
        """
        Create the space of permutations of a base string.

        :param base_string: an iterable of integer to denote the base string
        :raises TypeError: if one of the parameters has the wrong type
        :raises ValueError: if the parameters have the wrong value
        """
        if not isinstance(base_string, Iterable):
            raise type_error(base_string, "base_string", Iterable)

        string: Final[list[int]] = sorted(base_string)
        total: Final[int] = len(string)
        if total <= 0:
            raise ValueError(
                f"base string must not be empty, but is {base_string}.")

        # get data ranges
        self.__shape: Final[dict[int, int]] = {}
        minimum: int = string[0]
        maximum: int = string[0]
        for i in string:
            if not isinstance(i, int):
                raise type_error(i, "element of base_string", int)
            self.__shape[i] = self.__shape.get(i, 0) + 1
            if i < minimum:
                minimum = i
            if i > maximum:
                maximum = i

        # checking that the data is not empty
        different: Final[int] = len(self.__shape)
        if different <= 1:
            raise ValueError(
                "base_string must contain at least two different "
                f"elements, but is {base_string}.")

        # creating super class
        super().__init__(total, minimum, maximum)

        # start book
        #: a numpy array of the right type with the base string
        self.blueprint: Final[numpy.ndarray] = \
            numpy.array(string, dtype=self.dtype)
        # end book

        npoints: int = factorial(total)
        rep: int | None = self.__shape.get(minimum)
        for v in self.__shape.values():
            npoints = npoints // factorial(v)
            if v != rep:
                rep = None
        #: the exact number of different permutations
        self.__n_points = npoints

        #: the number of repetitions if all elements occur as same
        #: as often, or None otherwise
        self.__repetitions: Final[int | None] = rep

    def has_repetitions(self) -> bool:
        """
        Return whether elements occur repeatedly in the base string.

        :returns: `True` if at least one element occurs more than once,
            `False` otherwise
        """
        return (self.__repetitions is None) or (self.__repetitions > 1)

    def n(self) -> int:
        """
        Get the number of different values in the base string.

        :returns: the number of different values in the base string
        """
        return len(self.__shape)

    def is_dense(self) -> bool:
        """
        Check if all values in `min..max` appear in the permutation.

        :returns: `True` if the permutation is dense in the sense that
            all values from `self.min_value` to `self.max_value` appear.
        """
        return len(self.__shape) == (self.max_value - self.min_value + 1)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this space to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)

        reps: Final[int | None] = self.__repetitions
        if reps:
            logger.key_value(KEY_REPETITIONS, reps)
            if self.is_dense():
                return
        logger.key_value(KEY_BASE_STRING,
                         array_to_str(self.blueprint))

    def create(self) -> numpy.ndarray:  # +book
        r"""
        Create a permutation equal to the base string.

        The result is of the form [0, 0, 1, 1, 1, 2, 2...].

        :return: the permutation of the base string

        >>> perm = Permutations([1, 5, 2, 2, 4, 3, 4])
        >>> x = perm.create()
        >>> print(perm.to_str(x))
        1;2;2;3;4;4;5
        """
        return self.blueprint.copy()  # Create copy of the blueprint. # +book

    def validate(self, x: numpy.ndarray) -> None:
        """
        Validate a permutation of the base string.

        :param x: the integer string
        :raises TypeError: if the string is not an element of this space.
        :raises ValueError: if the shape of the vector is wrong or any of its
            element is not finite.
        """
        super().validate(x)
        counts: dict[int, int] = {}
        for xx in x:
            counts[xx] = counts.get(xx, 0) + 1

        if counts != self.__shape:
            for key in sorted(set(counts.keys()).union(
                    set(self.__shape.keys()))):
                exp = self.__shape.get(key, 0)
                found = counts.get(key, 0)
                if found != exp:
                    raise ValueError(
                        f"expected to find {key} exactly {exp} times "
                        f"but found it {found} times.")

    def n_points(self) -> int:
        """
        Get the number of possible different permutations of the base string.

        :return: factorial(simension) / Product(factorial(count(e)) for all e)

        >>> print(Permutations([0, 1, 2, 3, 0, 1, 2, 3]).n_points())
        2520
        """
        return self.__n_points

    def __str__(self) -> str:
        """
        Get the name of this space.

        :return: "perm" + blueprint string

        >>> print(Permutations([0, 1, 0, 2, 1]))
        perm_of_string
        >>> print(Permutations([0, 2, 0, 1, 1, 2]))
        perm3w2r
        >>> print(Permutations([0, 2, 1, 3]))
        perm4
        >>> print(Permutations([3, 2, 4, 2, 3, 2,4, 3, 4]))
        perm2to4w3r
        """
        minimum: Final[int] = self.min_value
        maximum: Final[int] = self.max_value
        reps: Final[int | None] = self.__repetitions
        different: Final[int] = self.n()
        if reps and (different != (self.dimension // reps)):
            raise ValueError(f"huh? {different} != {self.dimension} / {reps}")
        all_values: Final[bool] = self.is_dense()
        min_is_0: Final[bool] = minimum == 0

        if reps:
            if reps == 1:
                if all_values:
                    if min_is_0:
                        return f"perm{different}"
                    return sanitize_name(f"perm{minimum}to{maximum}")
            else:  # repetitions != 1
                if all_values:
                    if min_is_0:
                        return f"perm{different}w{reps}r"
                    return sanitize_name(f"perm{minimum}to{maximum}w{reps}r")
        return "perm_of_string"

    @staticmethod
    def standard(n: int) -> "Permutations":
        """
        Create a space for permutations of 0..n-1.

        :param n: the range of the values
        :returns: the permutations space
        """
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        if n <= 1:
            raise ValueError(f"n must be >1, but is {n}.")
        return Permutations(range(n))

    @staticmethod  # +book
    def with_repetitions(n: int, repetitions: int) -> "Permutations":  # +book
        """
        Create a space for permutations of `0..n-1` with repetitions.

        :param n: the range of the values
        :param repetitions: how often each value occurs
        :returns: the permutations space
        """
        if not isinstance(repetitions, int):
            raise type_error(repetitions, "repetitions", int)
        if repetitions < 1:
            raise ValueError(f"repetitions must be >0, but is {repetitions}.")
        if repetitions <= 1:
            return Permutations.standard(n)

        if not isinstance(n, int):
            raise type_error(n, "n", int)
        if n <= 1:
            raise ValueError(f"n must be >1, but is {n}.")
        return Permutations(list(range(n)) * repetitions)  # +book



"""A factory for functions checking whether argument values are new."""
from typing import Any, Callable


def is_new() -> Callable[[str], bool]:
    """
    Create a function returning `True` when seeing new values.

    Creates a function which returns `True` only the first time it receives a
    given argument and `False` all subsequent times.
    This is based on https://stackoverflow.com/questions/27427067

    :returns: a function `is_new(xx)` that will return `True` the first time
        it encounters any value `xx` and `False` for all values it has already
        seen

    >>> from moptipy.utils.cache import is_new
    >>> check = is_new()
    >>> print(check("a"))
    True
    >>> print(check("a"))
    False
    >>> print(check("b"))
    True
    >>> print(check("b"))
    False
    """
    s: dict[Any, int] = {}
    setdefault = s.setdefault
    n = 0  # noqa

    def add(x) -> bool:
        nonlocal n
        n += 1
        return setdefault(x, n) == n

    return add



"""
The set of selection algorithms.

:class:`~moptipy.algorithms.modules.selection.Selection` algorithms are
algorithms that choose elements from a pool of records based on their
:attr:`~moptipy.algorithms.modules.selection.FitnessRecord.fitness` and
random numbers. They are commonly used in Evolutionary Algorithms
(:class:`~moptipy.algorithms.so.general_ea.GeneralEA`).

Currently, the following selection algorithms have been implemented:

- :class:`~moptipy.algorithms.modules.selections.best.Best` selection picks
  the `n` best solutions from an array. It is the survival selection scheme
  used in a (mu+lambda) :class:`~moptipy.algorithms.so.ea.EA`.
- :class:`~moptipy.algorithms.modules.selections.random_without_replacement\
.RandomWithoutReplacement` randomly chooses `n` solutions without replacement.
  This is the mating selection scheme used in a (mu+lambda)
  :class:`~moptipy.algorithms.so.ea.EA`.
- :class:`~moptipy.algorithms.modules.selections.fitness_proportionate_sus\
.FitnessProportionateSUS` performs fitness proportionate selection for
  minimization using Stochastic Uniform Sampling. It also allows you to
  specify the selection probability for the worst element.
- :class:`~moptipy.algorithms.modules.selections.tournament.Tournament`
  performs tournament selection with a specified tournament size either with
  or without repetition.
"""



"""
The API for implementing optimization algorithms and -problems.

This package provides two things:

* the basic, abstract API for implementing optimization algorithms and
  problems
* the abstraction and implementation of black-box processes in form of
  :py:class:`~moptipy.api.process.Process` and its implementations

The former helps us to implement and plug together different components
of optimization problems and optimization algorithms.
The latter allows us to apply algorithms to problems and to collect the
results in a transparent way.
It also permits logging of the algorithm progress or even the change of
dynamic parameters.
"""



"""
Fitness Proportionate Selection with Stochastic Uniform Sampling.

In Fitness Proportionate Selection, the chance of a solution for being
selected is proportional to its fitness. This selection scheme is designed for
*maximization*, whereas all optimization in `moptipy` is done as
*minimization*. Therefore, some adjustments are necessary. We will discuss
them later. Let us first introduce the idea of fitness proportionate selection
for maximization.

This idea goes back to the original Genetic Algorithm by Holland. Let us say
that there are `N` elements and the element at index `i` has fitness `v(i)`.
The probability to select this element is then
`P(i) = v(i) / [sum_j=0^N v(j)]`, i.e., the fitness of the element divided by
the overall fitness sum. Let's say we have a population with the fitnesses
`1, 2, 3, 4`. The probability of each element of being selected is then the
fitness divided by the overall fitness sum (here `1+2+3+4=10`), i.e., we get
the probabilities `1/10=0.1, 0.2, 0.3, 0.4`. These nicely add up to `1`.

This can be implemented as follows: First, we copy the fitness values of the
individuals to an array `a` and get, in the above example, `a = [1, 2, 3, 4]`.
Then, we turn this array into a cumulative sum, i.e., add to each element the
sum of all previous elements. We get `a = [1, 3, 6, 10]`. `10`, the last
value, is the overall sum `S` of all fitnesses. Whenever we want to select an
element, we draw a random number `r` uniformly distributed in `[0, S)`. We
perform a binary search to get the right-most insertion index `i` of `r` in
`a`, i.e., the index `i` with `a[i - 1] <= r < a[i]`. Let's say you draw
`0.5`, then `i=0`, for `r=1` you get `i=1`, and for `r=9`, you get `i=3`. `i`
is then the element that has been selected.

In the classical Roulette Wheel Selection as used in Holland's original
Genetic Algorithm, we perform this sampling procedure (draw a random number
`r`, find the index `i` where it belongs, and return the corresponding
element) for each of the `n` offspring we want to sample. An alternative to
this is to perform Stochastic Universal Sampling (SUS) by Baker. Here, the
idea is that we only generate a single random number `r(0)` from within
`[0, S)`. The next number `r(1)` will not be random, but
`r(1)=(r(0) + S/n) mod S` and `r(i) = (r(i-1) + S/n) mod S` where `mod` be
the modulo division. In other words, after drawing the initial random sample,
we take steps of equal length along the Roulette Wheel, or, alternatively,
we have a Roulette Wheel not with a single choice point that we spin `n`
times but a wheel with `n` choice points that we spin a single time. This
avoids random jitter and also requires us to only draw a single random number.
We implement the fitness proportionate selection with stochastic uniform
sampling here.

But this is for *maximization*  of fitness, while we conduct *minimization*.

At first glance, we can turn a maximization problem into a minimization
problem by simply subtracting each fitness value from the maximum fitness
value `max_{all i} v(i)`. However, this has *two* big downsides.

Let us take the same example as before, the fitnesses `1, 2, 3, 4`. Under
maximization, the fitness `4` was best but now it is worst. It is also the
maximum fitness. So let us emulate fitness proportionate selection for
minimization by simply subtracting each fitness from the maximum (here: `4`).
We would get the adjusted fitnesses `3, 2, 1, 0`, the fitness sum `6`, and
probabilities `1/2, 1/3, 1/6, 0`, which still add up to `1` nicely. However,
now we have one element with 0% chance of being selected, namely the one with
the worst original fitness `4`. This is strange, because under maximization,
the worst element was `1` and it still had non-zero probability of being
selected. Furthermore, a second problem occurs if all elements are the same,
say, `1, 1, 1, 1`. We would then adjust them to all zeros, have a zero sum,
and getting `0/0` as selection probabilities. So that is not permissible.

The second problem can easily be solved by simply defining that, if all `N`
elements are identical, they should all get the same probability `1/N` of
being selected.

The first problem becomes clearer when we realize that under "normal" fitness
proportionate selection for maximization, the reference point "`0`" is
actually arbitrarily chosen and hard coded. If we have `1, 2, 3, 4, 10` as
fitnesses and transform them to the probabilities `0.05, 0.1, 0.15, 0.2, 0.5`,
we do so implicitly based on their "distance" to `0`. If we would add
some offset to them, say, `1`, i.e., calculate wit `2, 3, 4, 5, 11`, we would
get the fitness sum `25` and compute probabilities `0.08`, `0.12`, `0.16`,
`0.2`, and `0.44` instead. In other words, if we choose different reference
points, e.g., `-1` instead of `0`, we get different probabilities. And while
`0` seems a natural choice as reference point, it is actually just arbitrary.
The only actual condition of a reference point for maximization is that it
must be less or equal than/to the smallest occurring fitness.

If we do minimization instead of maximization, we do not have a "natural"
reference point. The only condition for the reference point is that it must be
larger or equal than/to the largest occurring fitness. Choosing the maximum
fitness value is just an arbitrary choice and it results in the solution of
this fitness getting `0` chance to reproduce.

If we can choose an arbitrary reference point for minimization, how do we
choose it? Our :class:`~moptipy.algorithms.modules.selections.\
fitness_proportionate_sus.FitnessProportionateSUS` has a parameter
`min_prob`, which corresponds to the minimal selection probability that *any*
element should have (of course, if we have `N` elements in the population, it
must hold that `0 <= min_prob < 1/N`). Based on this probability, we compute
the offset of the fitness values. We do it as follows:

The idea is that, in maximization, we got `P(i) = v(i) / [sum_j=0^N v(j)]`.
Now if `v(i) = 0`, we would get `P(i) = 0` as well. But we want
`P(i) = min_prob`, so we need to add an `offset` to each `v(i)`. So this then
becomes `P(i) = min_prob = offset / [sum_j=0^N (v(j) + offset)]`, which
becomes `min_prob = offset / [N * offset + sum_j=0^N v(j)]`. Let's set
`S = sum_j=0^N v(j)` to make this easier to read and we get
`min_prob = offset / (N * offset + S)`. Solving for `offset` gives us
`offset = (min_prob / (1.0 - (min_prob * N)))`. In other words, for any
allowable minimum selection probability `0<=min_prob<1/N`, we can compute an
offset to add to each fitness value that will result in the worst solution
having exactly this selection probability. The probabilities of the other
solutions will be larger, rather proportional to their fitness.

For minimization, first, we compute the maximum (i.e., worst) fitness
`max_fitness` and negate each fitness by subtracting it from `max_fitness`.
For an input array `1, 2, 3, 4` we now get `3, 2, 1, 0`. `S` be the sum of
the negated fitnesses, so in the above example, `S = 3 + 2 + 1 + 0 = 6`. We
can now compute the `offset` to be added to each negated fitness to achieve
the goal probability distribution as follows:
`offset = S * (min_prob / (1.0 - (min_prob * N)))`. If we had chosen
`min_prob = 0`, then `offset = 0` and the probability for the worst element
to be selected remains `0`. If we choose `min_prob = 0.01`, then we would
get `offset = 0.0625`. The selection probability of the worst element with
original fitness `4` and adjusted fitness `0` would be
`(0 + 0.0625) / (6 + (4 * 0.0625)) = 0.0625 / 6.25 = 0.01`.

As a side-effect of this choice of dynamic offsetting, our fitness
proportionate selection scheme becomes invariant under all translations of the
objective function value. The original fitness proportionate selection
schemes, regardless of being of the Roulette Wheel or Stochastic Universal
Sampling variant, do not have this property (see, for instance, de la Maza and
Tidor).

1. John Henry Holland. *Adaptation in Natural and Artificial Systems: An
   Introductory Analysis with Applications to Biology, Control, and Artificial
   Intelligence.* Ann Arbor, MI, USA: University of Michigan Press. 1975.
   ISBN: 0-472-08460-7
2. David Edward Goldberg. *Genetic Algorithms in Search, Optimization, and
   Machine Learning.* Boston, MA, USA: Addison-Wesley Longman Publishing Co.,
   Inc. 1989. ISBN: 0-201-15767-5
3. James E. Baker. Reducing Bias and Inefficiency in the Selection Algorithm.
   In John J. Grefenstette, editor, *Proceedings of the Second International
   Conference on Genetic Algorithms on Genetic Algorithms and their
   Application (ICGA'87),* July 1987, Cambridge, MA, USA, pages 14-21.
   Hillsdale, NJ, USA: Lawrence Erlbaum Associates. ISBN: 0-8058-0158-8
4. Peter J. B. Hancock. An Empirical Comparison of Selection Methods in
   Evolutionary Algorithms. In Terence Claus Fogarty, editor, *Selected Papers
   from the AISB Workshop on Evolutionary Computing (AISB EC'94),* April
   11-13, 1994, Leeds, UK, volume 865 of Lecture Notes in Computer Science,
   pages 80-94, Berlin/Heidelberg, Germany: Springer, ISBN: 978-3-540-58483-4.
   https://dx.doi.org/10.1007/3-540-58483-8_7. Conference organized by the
   Society for the Study of Artificial Intelligence and Simulation of
   Behaviour (AISB).
5. Tobias Blickle and Lothar Thiele. A Comparison of Selection Schemes used in
   Genetic Algorithms. Second edition, December 1995. TIK-Report 11 from the
   Eidgenössische Technische Hochschule (ETH) Zürich, Department of Electrical
   Engineering, Computer Engineering and Networks Laboratory (TIK), Zürich,
   Switzerland. ftp://ftp.tik.ee.ethz.ch/pub/publications/TIK-Report11.ps
6. Uday Kumar Chakraborty and Kalyanmoy Deb and Mandira Chakraborty. Analysis
   of Selection Algorithms: A Markov Chain Approach. *Evolutionary
   Computation,* 4(2):133-167. Summer 1996. Cambridge, MA, USA: MIT Press.
   doi:10.1162/evco.1996.4.2.133.
   https://dl.acm.org/doi/pdf/10.1162/evco.1996.4.2.133
7. Michael de la Maza and Bruce Tidor. An Analysis of Selection Procedures
   with Particular Attention Paid to Proportional and Bolzmann Selection. In
   Stephanie Forrest, editor, *Proceedings of the Fifth International
   Conference on Genetic Algorithms (ICGA'93),* July 17-21, 1993,
   Urbana-Champaign, IL, USA, pages 124-131. San Francisco, CA, USA:
   Morgan Kaufmann Publishers Inc. ISBN: 1-55860-299-2
"""

from math import isfinite
from typing import Any, Callable, Final

import numba  # type: ignore
import numpy as np
from numpy.random import Generator

from moptipy.algorithms.modules.selection import FitnessRecord, Selection
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_FLOAT
from moptipy.utils.strings import num_to_str_for_name
from moptipy.utils.types import type_error


@numba.njit(nogil=True, cache=True)
# start book
def _make_cum_sum(a: np.ndarray, offset_mul: float) -> None:
    """
    Compute the roulette wheel based on a given offset multiplier.

    The roulette wheel is basically an array of increasing values which
    corresponds to the cumulative sums of *"maximum - a[i] adjusted with
    the probability offset"*.

    :param a: the array with the fitness values
    :param offset_mul: the offset multiplier

    >>> import numpy as nn
    >>> ar = nn.array([1, 2, 3, 4], float)
    >>> _make_cum_sum(ar, 0)
    >>> list(map(str, ar))
    ['3.0', '5.0', '6.0', '6.0']
    >>> ar = nn.array([1, 2, 3, 4], float)
    >>> min_prob = 0.01
    >>> offset_mult = (min_prob / (1.0 - (min_prob * len(ar))))
    >>> _make_cum_sum(ar, offset_mult)
    >>> list(map(str, ar))
    ['3.0625', '5.125', '6.1875', '6.25']
    >>> (ar[-1] - ar[-2]) / ar[-1]  # compute prob of 4 being selected
    0.01
    >>> ar.fill(12)
    >>> _make_cum_sum(ar, 0.01)
    >>> list(map(str, ar))
    ['1.0', '2.0', '3.0', '4.0']
    """
    max_fitness: float = -np.inf
    min_fitness: float = np.inf
    for v in a:  # get minimum nd maximum fitness
        if v > max_fitness:
            max_fitness = v
        if v < min_fitness:
            min_fitness = v

    if min_fitness >= max_fitness:  # all elements are the same
        for i in range(len(a)):  # pylint: disable=C0200
            a[i] = i + 1  # assign equal probability
        return  # to all elements: each range = 1

    for i, v in enumerate(a):  # negate the array for minimization
        a[i] = max_fitness - v

    fitness_sum: Final[float] = a.sum()  # get the new fitness sum

    # compute the offset to accommodate the probability adjustment
    offset: Final[float] = fitness_sum * offset_mul

    cum_sum: float = 0.0
    for i, v in enumerate(a):  # compute the scaled cumulative sum
        a[i] = cum_sum = cum_sum + offset + v
# end book


# start book
class FitnessProportionateSUS(Selection):
    """Fitness Proportionate Selection with Stochastic Universal Sampling."""

# end book
    def __init__(self, min_prob: float = 0.0) -> None:
        """
        Create the stochastic universal sampling method.

        :param min_prob: the minimal selection probability of any element
        """
        super().__init__()
        if not isinstance(min_prob, float):
            raise type_error(min_prob, "min_prob", float)
        if not (0.0 <= min_prob < 0.2):
            raise ValueError(
                f"min_prob={min_prob}, but must be 0<=min_prob<0.2")
        #: the minimum selection probability of any element
        self.min_prob: Final[float] = min_prob
        #: the array to store the cumulative sum
        self.__cumsum: np.ndarray = np.empty(0, DEFAULT_FLOAT)

# start book
    def select(self, source: list[FitnessRecord],
               dest: Callable[[FitnessRecord], Any],
               n: int, random: Generator) -> None:
        """
        Perform deterministic best selection without replacement.

        :param source: the list with the records to select from
        :param dest: the destination collector to invoke for each selected
            record
        :param n: the number of records to select
        :param random: the random number generator
        """
        m: Final[int] = len(source)
        # compute the offset multiplier from the minimum probability
        # for this, min_prob must be < 1 / m
        min_prob: Final[float] = self.min_prob
        if min_prob >= (1.0 / m):  # -book
            raise ValueError(f"min_prob={min_prob} >= {1 / m}!")  # -book
        offset_mul: Final[float] = (min_prob / (1.0 - (min_prob * m)))
        # end book
        if (offset_mul < 0.0) or (not isfinite(offset_mul)):
            raise ValueError(
                f"min_prob={min_prob}, len={m} => offset_mul={offset_mul}")

        # start book
        a: np.ndarray = self.__cumsum  # get array for cumulative sum
        # end book
        if len(a) != m:  # re-allocate only if lengths don't match
            self.__cumsum = a = np.empty(m, DEFAULT_FLOAT)
        # start book
        for i, rec in enumerate(source):
            a[i] = rec.fitness  # store the fitnesses in the numpy array

        _make_cum_sum(a, offset_mul)  # construct cumulative sum array
        cum_sum: Final[float] = a[-1]  # cumulative sum = last element

        # stochastic uniform sampling
        current: float = random.uniform(0, cum_sum)  # starting point
        step_width: Final[float] = cum_sum / n  # step width

        for _ in range(n):  # select the `n` solutions
            dest(source[a.searchsorted(current, "right")])  # select
            current = (current + step_width) % cum_sum  # get next point
# end book

    def __str__(self):
        """
        Get the name of the stochastic uniform sampling selection algorithm.

        :return: the name of the stochastic uniform sampling selection
            algorithm
        """
        return "fpsus" if self.min_prob <= 0.0 \
            else f"fpsus{num_to_str_for_name(self.min_prob)}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("minprob", self.min_prob, also_hex=True)



"""
The base classes for implementing optimization algorithms.

All optimization algorithms implemented based on the `moptipy` API inherit
from :class:`~moptipy.api.algorithm.Algorithm`. If you implement a new
algorithm, you can test with the pre-defined unit test routine
:func:`~moptipy.tests.algorithm.validate_algorithm`.
"""
from typing import Final

from moptipy.api.component import Component
from moptipy.api.logging import SCOPE_OP0, SCOPE_OP1, SCOPE_OP2
from moptipy.api.operators import (
    Op0,
    Op1,
    Op2,
    check_op0,
    check_op1,
    check_op2,
)
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import PART_SEPARATOR
from moptipy.utils.types import type_error


# start book
class Algorithm(Component):
    """A base class for implementing optimization algorithms."""

    def solve(self, process: Process) -> None:
        """
        Apply this optimization algorithm to the given process.

        :param process: the process which provides methods to access the
            search space, the termination criterion, and a source of
            randomness. It also wraps the objective function, remembers the
            best-so-far solution, and takes care of creating log files (if
            this is wanted).
        """
# end book


class Algorithm0(Algorithm):
    """An algorithm with a nullary search operator."""

    def __init__(self, name: str, op0: Op0) -> None:
        """
        Create the algorithm with nullary search operator.

        :param name: the name of the algorithm
        :param op0: the nullary search operator
        """
        #: The nullary search operator.
        self.op0: Final[Op0] = check_op0(op0)
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        if len(name) <= 0:
            raise ValueError(f"Algorithm name cannot be '{name}'.")
        #: the name of this optimization algorithm, which is also the return
        #: value of `__str__()`
        self.name: Final[str] = name

    def __str__(self) -> str:
        """
        Get the name of the algorithm.

        :return: the name of the algorithm
        """
        return self.name

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_OP0) as sc:
            self.op0.log_parameters_to(sc)


class Algorithm1(Algorithm0):
    """An algorithm with a unary search operator."""

    def __init__(self, name: str, op0: Op0, op1: Op1) -> None:
        """
        Create the algorithm with nullary and unary search operator.

        :param name: the name of the algorithm
        :param op0: the nullary search operator
        :param op1: the unary search operator
        """
        super().__init__(name if op1.__class__ == Op1 else
                         f"{name}{PART_SEPARATOR}{op1}", op0)
        #: The unary search operator.
        self.op1: Final[Op1] = check_op1(op1)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_OP1) as sc:
            self.op1.log_parameters_to(sc)


class Algorithm2(Algorithm1):
    """An algorithm with a binary and unary operator."""

    def __init__(self, name: str, op0: Op0, op1: Op1, op2: Op2) -> None:
        """
        Create the algorithm with nullary, unary, and binary search operator.

        :param name: the name of the algorithm
        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param op2: the binary search operator
        """
        super().__init__(
            name if op2.__class__ is Op2 else
            f"{name}{PART_SEPARATOR}{op2}", op0, op1)
        #: The binary search operator.
        self.op2: Final[Op2] = check_op2(op2)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_OP2) as sc:
            self.op2.log_parameters_to(sc)


def check_algorithm(algorithm: Algorithm) -> Algorithm:
    """
    Check whether an object is a valid instance of :class:`Algorithm`.

    :param algorithm: the algorithm object
    :return: the object
    :raises TypeError: if `algorithm` is not an instance of :class:`Algorithm`
    """
    if not isinstance(algorithm, Algorithm):
        raise type_error(algorithm, "algorithm", Algorithm)
    return algorithm



"""
A multi-objective version of the Randomized Local Search algorithm.

The original randomized local search (RLS) always remembers the best-so-far
solution and, in each step, generates a new one by applying the unary search
operator. If the new solution is not worse than the best-so-far one (according
to the single objective function), it becomes the new best-so-far solution.
In this multi-objective version, we accept the new solution if it is not
dominated by the best-so-far solution.
"""
from typing import Callable, Final

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm1
from moptipy.api.mo_algorithm import MOAlgorithm
from moptipy.api.mo_process import MOProcess
from moptipy.api.operators import Op0, Op1


class MORLS(Algorithm1, MOAlgorithm):
    """The MO-RLS is a local search accepting all non-worsening moves."""

    def solve_mo(self, process: MOProcess) -> None:
        """
        Apply the MO-RLS to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        best_x = process.create()  # record for best-so-far solution
        best_f = process.f_create()  # the objective values
        new_x = process.create()  # record for new solution
        new_f = process.f_create()  # the objective values
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.f_evaluate  # the objective
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate
        domination: Final[Callable] = process.f_dominates

        # Start at a random point in the search space and evaluate it.
        self.op0.op0(random, best_x)  # Create 1 solution randomly and
        evaluate(best_x, best_f)  # evaluate it.

        while not should_terminate():  # Until we need to quit...
            op1(random, new_x, best_x)  # new_x = neighbor of best_x
            evaluate(new_x, new_f)
            if domination(new_f, best_f) <= 0:  # new is not worse than best?
                best_f, new_f = new_f, best_f  # swap objective values.
                best_x, new_x = new_x, best_x  # swap best and new.

        process.check_in(best_x, best_f)  # check-in final result

    def __init__(self, op0: Op0, op1: Op1) -> None:
        """
        Create the randomized local search (rls).

        :param op0: the nullary search operator
        :param op1: the unary search operator
        """
        Algorithm1.__init__(self, "morls", op0, op1)



"""The `logger` routine for writing a log string to stdout."""
import datetime
from contextlib import AbstractContextManager, nullcontext
from typing import Callable, Final

#: the "now" function
__DTN: Final[Callable[[], datetime.datetime]] = datetime.datetime.now


def logger(message: str, note: str = "",
           lock: AbstractContextManager = nullcontext()) -> None:
    """
    Write a message to the log.

    :param message: the message
    :param note: a note to put between the time and the message
    :param lock: the lock to prevent multiple threads to write log
        output at the same time
    """
    text: Final[str] = f"{__DTN()}{note}: {message}"
    with lock:
        print(text, flush=True)  # noqa



"""Functions that can be used to test algorithm implementations."""
from math import inf, isfinite
from typing import Any, Final

from moptipy.api.algorithm import (
    Algorithm,
    Algorithm0,
    Algorithm1,
    Algorithm2,
    check_algorithm,
)
from moptipy.api.encoding import Encoding
from moptipy.api.execution import Execution
from moptipy.api.objective import Objective
from moptipy.api.operators import check_op0, check_op1, check_op2
from moptipy.api.space import Space
from moptipy.tests.component import validate_component
from moptipy.tests.encoding import validate_encoding
from moptipy.tests.objective import validate_objective
from moptipy.tests.space import validate_space
from moptipy.utils.types import type_error


def validate_algorithm(algorithm: Algorithm,
                       solution_space: Space,
                       objective: Objective,
                       search_space: Space | None = None,
                       encoding: Encoding | None = None,
                       max_fes: int = 100,
                       required_result: int | float | None = None,
                       uses_all_fes_if_goal_not_reached: bool = True,
                       is_encoding_deterministic: bool = True) \
        -> None:
    """
    Check whether an algorithm follows the moptipy API specification.

    :param algorithm: the algorithm to test
    :param solution_space: the solution space
    :param objective: the objective function
    :param search_space: the optional search space
    :param encoding: the optional encoding
    :param max_fes: the maximum number of FEs
    :param required_result: the optional required result quality
    :param uses_all_fes_if_goal_not_reached: will the algorithm use all FEs
        unless it reaches the goal?
    :param is_encoding_deterministic: is the encoding deterministic?
    :raises TypeError: if `algorithm` is not a
        :class:`~moptipy.api.algorithm.Algorithm` instance
    :raises ValueError: if `algorithm` does not behave like it should
    """
    if not isinstance(algorithm, Algorithm):
        raise type_error(algorithm, "algorithm", Algorithm)

    check_algorithm(algorithm)
    if isinstance(algorithm, Algorithm0):
        check_op0(algorithm.op0)
        if isinstance(algorithm, Algorithm1):
            check_op1(algorithm.op1)
            if isinstance(algorithm, Algorithm2):
                check_op2(algorithm.op2)

    validate_component(algorithm)
    validate_space(solution_space, None)
    validate_objective(objective, None, None)
    if encoding is not None:
        validate_encoding(encoding, None, None, None,
                          is_encoding_deterministic)
        validate_space(search_space, None)

    if not isinstance(max_fes, int):
        raise type_error(max_fes, "max_fes", int)
    if max_fes <= 0:
        raise ValueError(f"max_fes must be > 0, but is {max_fes} "
                         f"for {algorithm} on {objective}")

    exp = Execution()
    exp.set_algorithm(algorithm)
    exp.set_max_fes(max_fes)
    exp.set_solution_space(solution_space)
    exp.set_objective(objective)
    if search_space is not None:
        exp.set_search_space(search_space)
        exp.set_encoding(encoding)

    lb: Final[int | float] = objective.lower_bound()
    if (not isfinite(lb)) and (lb != -inf):
        raise ValueError(f"objective lower bound cannot be {lb}"
                         f" for {algorithm} on objective {objective}.")
    ub = objective.upper_bound()
    if (not isfinite(ub)) and (ub != inf):
        raise ValueError(f"objective upper bound cannot be {ub}"
                         f" for {algorithm} on objective {objective}.")

    goal: int | float = lb
    if required_result is not None:
        if not (lb <= required_result <= ub):
            raise ValueError(f"required result must be in [{lb},{ub}], "
                             f"for {algorithm} on {objective} but "
                             f"is {required_result}")
        if (not isfinite(required_result)) and (required_result != -inf):
            raise ValueError(f"required_result must not be {required_result} "
                             f"for {algorithm} on {objective}.")
        goal = required_result
    exp.set_goal_f(goal)

    with exp.execute() as process:
        # re-raise any exception that was caught
        if hasattr(process, "_caught"):
            error = getattr(process, "_caught")
            if error is not None:
                raise error
        # no exception? ok, let's check the data
        if not process.has_best():
            raise ValueError(f"The algorithm {algorithm} did not produce "
                             f"any solution on {objective}.")

        if not process.should_terminate():
            if uses_all_fes_if_goal_not_reached:
                raise ValueError(f"The algorithm {algorithm} stopped before "
                                 "hitting the termination criterion "
                                 f"on {objective}.")

        consumed_fes: Final[int] = process.get_consumed_fes()
        if not isinstance(consumed_fes, int):
            raise type_error(consumed_fes, "consumed_fes", int)
        if (consumed_fes <= 0) or (consumed_fes > max_fes):
            raise ValueError(
                f"Consumed FEs must be positive and <= {max_fes}, "
                f"but is {consumed_fes} for {algorithm} on {objective}.")

        last_imp_fe: Final[int] = process.get_last_improvement_fe()
        if not isinstance(last_imp_fe, int):
            raise type_error(last_imp_fe, "last improvement FE", int)
        if (last_imp_fe <= 0) or (last_imp_fe > consumed_fes):
            raise ValueError("Last improvement FEs must be positive and "
                             f"<= {consumed_fes}, but is {last_imp_fe} "
                             f"for {algorithm} on {objective}.")

        consumed_time: Final[int] = process.get_consumed_time_millis()
        if not isinstance(consumed_time, int):
            raise type_error(consumed_time, "consumed time", int)
        if consumed_time < 0:
            raise ValueError(
                f"Consumed time must be >= 0, but is {consumed_time} "
                f"for {algorithm} on {objective}.")

        last_imp_time: Final[int] = process.get_last_improvement_time_millis()
        if not isinstance(last_imp_time, int):
            raise type_error(last_imp_time, "last improvement time", int)
        if (last_imp_time < 0) or (last_imp_time > consumed_time):
            raise ValueError(
                f"Consumed time must be >= 0 and <= {consumed_time}, but "
                f"is {last_imp_time} for {algorithm} on {objective}.")

        if lb != process.lower_bound():
            raise ValueError(
                "Inconsistent lower bounds between process "
                f"({process.lower_bound()}) and objective ({lb})"
                f" for {algorithm} on {objective}.")
        if ub != process.upper_bound():
            raise ValueError(
                "Inconsistent upper bounds between process "
                f"({process.upper_bound()}) and objective ({ub}) "
                f" for {algorithm} on {objective}.")

        res_f: Final[float | int] = process.get_best_f()
        if not isfinite(res_f):
            raise ValueError(f"Infinite objective value of result "
                             f"for {algorithm} on {objective}.")
        if (res_f < lb) or (res_f > ub):
            raise ValueError(f"Objective value {res_f} outside of bounds "
                             f"[{lb},{ub}] for {algorithm} on {objective}.")

        if required_result is not None:
            if res_f > required_result:
                raise ValueError(
                    f"Algorithm {algorithm} should find solution of quality "
                    f"{required_result} on {objective}, but got one of "
                    f"{res_f}.")

        if res_f <= goal:
            if last_imp_fe != consumed_fes:
                raise ValueError(
                    f"if result={res_f} is as good as goal={goal}, then "
                    f"last_imp_fe={last_imp_fe} must equal"
                    f" consumed_fe={consumed_fes} for {algorithm} on "
                    f"{objective}.")
            if (10_000 + (1.05 * last_imp_time)) < consumed_time:
                raise ValueError(
                    f"if result={res_f} is as good as goal={goal}, then "
                    f"last_imp_time={last_imp_time} must not be much less"
                    f" than consumed_time={consumed_time} for {algorithm} "
                    f"on {objective}.")
        else:
            if uses_all_fes_if_goal_not_reached and (consumed_fes != max_fes):
                raise ValueError(
                    f"if result={res_f} is worse than goal={goal}, then "
                    f"consumed_fes={consumed_fes} must equal "
                    f"max_fes={max_fes} for {algorithm} on {objective}.")

        y = solution_space.create()
        process.get_copy_of_best_y(y)
        solution_space.validate(y)
        check_f = objective.evaluate(y)
        if check_f != res_f:
            raise ValueError(
                f"Inconsistent objective value {res_f} from process compared "
                f"to {check_f} from objective function for {algorithm} on "
                f"{objective}.")

        x: Any | None = None
        if search_space is not None:
            x = search_space.create()
            process.get_copy_of_best_x(x)
            search_space.validate(x)

        if encoding is not None:
            y2 = solution_space.create()
            encoding.decode(x, y2)
            solution_space.validate(y2)
            if is_encoding_deterministic \
                    and not solution_space.is_equal(y, y2):
                raise ValueError(
                    f"error when mapping point in search space {x} to "
                    f"solution {y2}, because it should be {y} for "
                    f"{algorithm} on {objective} under encoding {encoding}")



"""Run the moptipy example experiment."""
import os.path as pp
import sys
from typing import Any, Callable, Final, Iterable, cast

import moptipy.api.experiment as ex
from moptipy.algorithms.modules.selections.fitness_proportionate_sus import (
    FitnessProportionateSUS,
)
from moptipy.algorithms.modules.selections.tournament import Tournament
from moptipy.algorithms.random_sampling import RandomSampling
from moptipy.algorithms.random_walk import RandomWalk
from moptipy.algorithms.single_random_sample import SingleRandomSample
from moptipy.algorithms.so.ea import EA
from moptipy.algorithms.so.fitnesses.direct import Direct
from moptipy.algorithms.so.fitnesses.rank import Rank
from moptipy.algorithms.so.general_ea import GeneralEA
from moptipy.algorithms.so.hill_climber import HillClimber
from moptipy.algorithms.so.hill_climber_with_restarts import (
    HillClimberWithRestarts,
)
from moptipy.algorithms.so.rls import RLS
from moptipy.api.algorithm import Algorithm
from moptipy.api.execution import Execution
from moptipy.examples.jssp.gantt_space import GanttSpace
from moptipy.examples.jssp.instance import Instance
from moptipy.examples.jssp.makespan import Makespan
from moptipy.examples.jssp.ob_encoding import OperationBasedEncoding
from moptipy.operators.permutations.op0_shuffle import Op0Shuffle
from moptipy.operators.permutations.op1_swap2 import Op1Swap2
from moptipy.operators.permutations.op1_swapn import Op1SwapN
from moptipy.operators.permutations.op2_gap import (
    Op2GeneralizedAlternatingPosition,
)
from moptipy.spaces.permutations import Permutations
from moptipy.utils.console import logger
from moptipy.utils.help import help_screen
from moptipy.utils.path import Path
from moptipy.utils.types import type_error

#: The default instances to be used in our experiment. These have been
#: computed via instance_selector.propose_instances.
#: The instances in this tuple are sorted by the scale of the search space,
#: if that search space is the order-based encoding, i.e., permutations with
#: repetitions. This is almost the same as an ordering by the number  of
#: possible (feasible or infeasible) Gantt charts than can be constructed, with
#: one exception: For `dmu67`, the search space size is `2.768*(10**1241)` and
#: for `dmu72`, it is `3.862*(10**1226)`, i.e., `dmu67` has a larger search
#: order-based encoding search space than `dmu72`. However, for `dmu72`, we can
#: construct `1.762*(10**967)` different (feasible or infeasible) Gantt charts,
#: whereas for `dmu67`, we can only construct `1.710*(10**958)`, i.e., fewer.
#: Therefore, `dmu67` in this ordering here comes after `dmu72`, but it would
#: come before if we were sorting instances by the solution space size.
EXPERIMENT_INSTANCES: \
    Final[tuple[str, str, str, str, str, str, str, str]] = \
    ("orb06", "la38", "abz8", "yn4", "swv14", "dmu72", "dmu67", "ta70")

#: The number of runs per instance in our JSSP experiment.
#: For each combination of algorithm setup and JSSP instance, we will
#: perform exactly this many runs.
EXPERIMENT_RUNS: Final[int] = 23

#: We will perform two minutes per run.
EXPERIMENT_RUNTIME_MS: Final[int] = 2 * 60 * 1000

#: The default set of algorithms for our experiments.
#: Each of them is a Callable that receives two parameters, the instance
#: `inst` and the permutation with repetitions-space `pwr`.
DEFAULT_ALGORITHMS: Final[list[
    Callable[[Instance, Permutations], Algorithm]]] = [
    lambda inst, pwr: SingleRandomSample(Op0Shuffle(pwr)),  # single sample
    lambda inst, pwr: RandomSampling(Op0Shuffle(pwr)),  # random sampling
    lambda inst, pwr: HillClimber(Op0Shuffle(pwr), Op1Swap2()),  # hill climb.
    lambda inst, pwr: RLS(Op0Shuffle(pwr), Op1Swap2()),  # RLS
    lambda inst, pwr: RandomWalk(Op0Shuffle(pwr), Op1Swap2()),  # random walk
    lambda inst, pwr: HillClimber(Op0Shuffle(pwr), Op1SwapN()),  # hill climb.
    lambda inst, pwr: RLS(Op0Shuffle(pwr), Op1SwapN()),  # RLS
]
for scale in range(7, 21):  # add the hill climbers with restarts
    DEFAULT_ALGORITHMS.append(cast(
        Callable[[Instance, Permutations], Algorithm],
        lambda inst, pwr, i=scale: HillClimberWithRestarts(
            Op0Shuffle(pwr), Op1Swap2(), 2 ** i)))  # hill clim. with restarts
    DEFAULT_ALGORITHMS.append(cast(
        Callable[[Instance, Permutations], Algorithm],
        lambda inst, pwr, i=scale: HillClimberWithRestarts(
            Op0Shuffle(pwr), Op1SwapN(), 2 ** i)))  # hill clim. with restarts
for log2_mu in range(0, 14):
    mu: int = 2 ** log2_mu
    for lambda_ in sorted({1, max(1, log2_mu), round(mu ** 0.5), mu, mu + mu,
                           4 * mu, 8 * mu}):
        if lambda_ > 65536:
            continue
        DEFAULT_ALGORITHMS.append(cast(
            Callable[[Instance, Permutations], Algorithm],
            lambda inst, pwr, mm=mu, ll=lambda_: EA(
                Op0Shuffle(pwr), Op1Swap2(), None, mm, ll, 0.0)))
        if (mu == lambda_) and (mu in {2, 4, 32, 256, 4096}):
            for br_exp in range(1, 11):
                DEFAULT_ALGORITHMS.append(cast(
                    Callable[[Instance, Permutations], Algorithm],
                    lambda inst, pwr, ml=mu, br=(2 ** -br_exp): EA(
                        Op0Shuffle(pwr), Op1Swap2(),
                        Op2GeneralizedAlternatingPosition(pwr), ml, ml, br)))
            for br_exp in range(2, 11):
                DEFAULT_ALGORITHMS.append(cast(
                    Callable[[Instance, Permutations], Algorithm],
                    lambda inst, pwr, ml=mu, br=1.0 - (2 ** -br_exp): EA(
                        Op0Shuffle(pwr), Op1Swap2(),
                        Op2GeneralizedAlternatingPosition(pwr), ml, ml, br)))

for mu_lambda in [4, 32]:
    DEFAULT_ALGORITHMS.append(cast(
        Callable[[Instance, Permutations], Algorithm],
        lambda inst, pwr, ml=mu_lambda: GeneralEA(
            Op0Shuffle(pwr), Op1Swap2(),
            Op2GeneralizedAlternatingPosition(pwr), ml, ml, 2 ** -3,
            survival=Tournament(2))))
    DEFAULT_ALGORITHMS.append(cast(
        Callable[[Instance, Permutations], Algorithm],
        lambda inst, pwr, ml=mu_lambda: GeneralEA(
            Op0Shuffle(pwr), Op1Swap2(),
            Op2GeneralizedAlternatingPosition(pwr), ml, ml, 2 ** -3,
            survival=Tournament(4))))
    DEFAULT_ALGORITHMS.append(cast(
        Callable[[Instance, Permutations], Algorithm],
        lambda inst, pwr, ml=mu_lambda: GeneralEA(
            Op0Shuffle(pwr), Op1Swap2(),
            Op2GeneralizedAlternatingPosition(pwr), ml, ml, 2 ** -3,
            mating=Tournament(2, replacement=False))))
    DEFAULT_ALGORITHMS.append(cast(
        Callable[[Instance, Permutations], Algorithm],
        lambda inst, pwr, ml=mu_lambda: GeneralEA(
            Op0Shuffle(pwr), Op1Swap2(),
            Op2GeneralizedAlternatingPosition(pwr), ml, ml, 2 ** -3,
            fitness=Direct(), survival=FitnessProportionateSUS())))
    DEFAULT_ALGORITHMS.append(cast(
        Callable[[Instance, Permutations], Algorithm],
        lambda inst, pwr, ml=mu_lambda: GeneralEA(
            Op0Shuffle(pwr), Op1Swap2(),
            Op2GeneralizedAlternatingPosition(pwr), ml, ml, 2 ** -3,
            fitness=Direct(), survival=FitnessProportionateSUS(),
            mating=Tournament(2, replacement=False))))
    DEFAULT_ALGORITHMS.append(cast(
        Callable[[Instance, Permutations], Algorithm],
        lambda inst, pwr, ml=mu_lambda: GeneralEA(
            Op0Shuffle(pwr), Op1Swap2(),
            Op2GeneralizedAlternatingPosition(pwr), ml, ml, 2 ** -3,
            fitness=Rank(), survival=FitnessProportionateSUS(),
            mating=Tournament(2, replacement=False))))


def run_experiment(base_dir: str = pp.join(".", "results"),
                   algorithms: Iterable[
                       Callable[[Instance, Permutations],
                                Algorithm]] = tuple(DEFAULT_ALGORITHMS),
                   instances: Iterable[str] = EXPERIMENT_INSTANCES,
                   n_runs: int = EXPERIMENT_RUNS,
                   max_time: int | None = EXPERIMENT_RUNTIME_MS,
                   max_fes: int | None = None,
                   n_threads: int | None = None) -> None:
    """
    Run the experiment.

    :param algorithms: the algorithm factories.
        Each factory receives as input one JSSP `Instance` and one instance
        of `PermutationWithRepetition`. It returns either an instance of
        `Algorithm` or of `Execution`.
    :param instances: the JSSP instance names
    :param base_dir: the base directory
    :param n_runs: the number of runs
    :param max_time: the maximum runtime in milliseconds
    :param max_fes: the maximum runtime in FEs
    :param n_threads: the number of threads
    """
    # The initial parameter validity checks.
    if not isinstance(base_dir, str):
        raise type_error(base_dir, "base_dir", str)
    if not isinstance(algorithms, Iterable):
        raise type_error(algorithms, "algorithms", Iterable)
    if not isinstance(instances, Iterable):
        raise type_error(instances, "instances", Iterable)
    if not isinstance(n_runs, int):
        raise type_error(n_runs, "n_runs", int)
    if n_runs <= 0:
        raise ValueError(f"n_runs must be positive, but is {n_runs}.")
    if max_time is not None:
        if not isinstance(max_time, int):
            raise type_error(max_time, "max_time", int)
        if max_time <= 0:
            raise ValueError(f"max_time must be positive, but is {max_time}.")
    if max_fes is not None:
        if not isinstance(max_fes, int):
            raise type_error(max_fes, "max_fes", int)
        if max_time <= 0:
            raise ValueError(f"max_fes must be positive, but is {max_fes}.")
    if (max_fes is None) and (max_time is None):
        raise ValueError("Either max_fes or max_time must be provided.")

    inst_gens = [(lambda s=s: Instance.from_resource(s)) for s in instances]
    if len(inst_gens) <= 0:
        raise ValueError("There must be at least one instance.")
    del instances

    # In this loop, we convert the simple algorithm lambdas to execution
    # creators.
    algo_gens = []
    for algo in algorithms:
        # we create one constructor for each algorithm factory
        def creator(inst: Instance, algor: Callable = algo) -> Execution:
            """
            Create the algorithm for a given instance.

            :param inst: the JSSP instance
            :param algor: the algorithm creator
            :return: an Execution for the experiment
            """
            pwr: Permutations = Permutations.with_repetitions(
                inst.jobs, inst.machines)

            val: Execution | Algorithm = algor(inst, pwr)
            experiment: Execution
            if isinstance(val, Execution):
                experiment = cast(Execution, val)
            else:
                if not isinstance(val, Algorithm):
                    raise type_error(val, "result of factory function",
                                     Algorithm)
                experiment = Execution()
                experiment.set_algorithm(cast(Algorithm, val))

            if max_time is not None:
                experiment.set_max_time_millis(max_time)
            if max_fes is not None:
                experiment.set_max_fes(max_fes)
            experiment.set_objective(Makespan(inst))
            experiment.set_solution_space(GanttSpace(inst))
            experiment.set_search_space(pwr)
            experiment.set_encoding(OperationBasedEncoding(inst))
            experiment.set_log_improvements()
            return experiment

        algo_gens.append(creator)  # add constructor to generator list
        del creator  # the creator is no longer needed
    del algorithms

    if len(algo_gens) <= 0:
        raise ValueError("There must be at least one algorithm.")

    ikwargs: dict[str, Any] = {"base_dir": base_dir,
                               "instances": inst_gens,
                               "setups": algo_gens,
                               "n_runs": n_runs,
                               "perform_warmup": True,
                               "warmup_fes": 20,
                               "perform_pre_warmup": True,
                               "pre_warmup_fes": 20}
    if n_threads is not None:
        ikwargs["n_threads"] = n_threads

    ex.run_experiment(**ikwargs)  # invoke the actual experiment


# Execute experiment if run as script
if __name__ == "__main__":
    help_screen(
        "JSSP Experiment Executor", __file__,
        "Run the JSSP experiment.",
        [("results_dir",
          "the directory where the results should be stored",
          True),
         ("n_threads",
          "the number of threads to use for the experiment",
          True)])
    mkwargs: dict[str, Any] = {}
    if len(sys.argv) > 1:
        dest_dir: Final[Path] = Path.path(sys.argv[1])
        dest_dir.ensure_dir_exists()
        mkwargs["base_dir"] = dest_dir
        logger(f"Set base_dir '{dest_dir}'.")
        if len(sys.argv) > 2:
            n_cpu = int(sys.argv[2])
            mkwargs["n_threads"] = n_cpu
            logger(f"Set n_threads to {n_cpu}.")

    run_experiment(**mkwargs)  # invoke the experiment execution



"""Provides statistical comparison tables."""

from math import isfinite
from statistics import mean, median
from typing import Any, Callable, Final, Iterable, cast

from scipy.stats import mannwhitneyu  # type: ignore

from moptipy.api.logging import KEY_BEST_F
from moptipy.evaluation.end_results import EndResult
from moptipy.utils.formatted_string import FormattedStr
from moptipy.utils.markdown import Markdown
from moptipy.utils.number_renderer import (
    DEFAULT_NUMBER_RENDERER,
    NumberRenderer,
)
from moptipy.utils.path import Path
from moptipy.utils.table import Table
from moptipy.utils.text_format import TextFormatDriver
from moptipy.utils.types import type_error

#: the string constant for alpha
__ALPHA: Final[FormattedStr] = FormattedStr.special("\u03b1")
#: the string constant for dash or unclear
__DASH: Final[FormattedStr] = FormattedStr.special("\u2014")
#: the character for less
__LESS: Final[FormattedStr] = FormattedStr("<", code=True)
#: the character for equal
__UNCLEAR: Final[FormattedStr] = FormattedStr("?", code=True)
#: the character for greater
__GREATER: Final[FormattedStr] = FormattedStr(">", code=True)

#: the name of the win-equal-loss column
__WEL: Final[tuple[FormattedStr, str, FormattedStr, str, FormattedStr]] \
    = __LESS, " / ", __UNCLEAR, " / ", __GREATER


def __compare(a: list[int | float],
              b: list[int | float]) -> int:
    """
    Compare two lists of numbers.

    `-1` is returned if `a` has both a smaller mean and median value than `b`.
    `1` is returned if `a` has both a larger mean and median value than `b`.
    `0` is returned otherwise.

    :param a: the first list of numbers
    :param b: the second list of numbers
    :returns: `-1` if the numbers in `a` tend to be smaller than those in `b`,
        `1` if the numbers in `a` tend to be larger than those in `b`,
        `0` if undecided
    """
    smaller: bool
    ma = mean(a)
    mb = mean(b)
    if ma < mb:
        smaller = True
    elif ma > mb:
        smaller = False
    else:
        return 0
    ma = median(a)
    mb = median(b)
    if ma < mb:
        return -1 if smaller else 0
    if ma > mb:
        return 0 if smaller else 1
    return 0


def __compare_to_str(cmp: int, p_str: FormattedStr | None,
                     p: float, alpha_prime: float) -> str | list[str]:
    """
    Transform a comparison result and a string to an output.

    :param cmp: the comparison result
    :param p_str: the rendered `p` value string
    :param p: the actual `p` value
    :param alpha_prime: the `alpha` value
    :returns: the output
    """
    if cmp == 0:
        return __DASH
    if p >= alpha_prime:
        return [p_str, " ", __UNCLEAR]
    if cmp < 0:
        return [p_str, " ", __LESS]
    if cmp > 0:
        return [p_str, " ", __GREATER]
    raise ValueError(
        f"huh? p={p}, p_str={p_str}, cmp={cmp}, alpha'={alpha_prime}")


def tabulate_result_tests(
        end_results: Iterable[EndResult],
        file_name: str = "tests",
        dir_name: str = ".",
        alpha: float = 0.02,
        text_format_driver: TextFormatDriver | Callable[[], TextFormatDriver]
        = Markdown.instance,
        algorithm_sort_key: Callable[[str], Any] = lambda a: a,
        instance_sort_key: Callable[[str], Any] = lambda i: i,
        instance_namer: Callable[[str], str] = lambda x: x,
        algorithm_namer: Callable[[str], str] = lambda x: x,
        use_lang: bool = False,
        p_renderer: NumberRenderer = DEFAULT_NUMBER_RENDERER,
        value_getter: Callable[[EndResult], int | float]
        = EndResult.getter(KEY_BEST_F)) -> Path:
    r"""
    Tabulate the results of statistical comparisons of end result qualities.

    `end_results` contains a sequence of
    :class:`~moptipy.evaluation.end_results.EndResult` records, each of which
    represents the result of one run of one algorithm on one instance. This
    function performs a two-tailed Mann-Whitney U test for each algorithm pair
    on each problem instance to see if the performances are statistically
    significantly different. The results of these tests are tabulated,
    together with their `p`-values, i.e., the probabilities that the observed
    differences would occur if the two algorithms would perform the same.

    If `p` is sufficiently small, this means that it is unlikely that the
    difference in performance of the two compared algorithms that was observed
    stems from randomness. But what does "sufficiently small" mean?
    As parameter, this function accepts a significance threshold
    `0<alpha<0.5`. `alpha` is, so to say, the upper limit of the "probability
    to be wrong" if we claim something like "algorithm A is better than
    algorithm B" that we are going to accept. In other words, if the table
    says that algorithm A is better than algorithm B, the chance that this is
    wrong is not more than `alpha`.

    However, if we do many such tests, our chance to make at least one mistake
    grows. If we do `n_tests` tests, then the chance that all of them are
    "right" would be `1-[(1-alpha)^n_tests]`. Since we are going to do
    multiple tests, the Bonferroni correction is therefore applied and
    `alpha'=alpha/n_tests` is computed. Then, the chance to have at least one
    of the `n_tests` test results to be wrong is not higher than `alpha`.

    The test results are presented as follows:
    The first column of the generated table denotes the problem instances.
    Each of the other columns represents a pair of algorithms. In each cell,
    the pair is compared based on the results on the instance of the row. The
    cell ten holds the `p`-value of the two-tailed Mann-Whitney U test. If the
    first algorithm is significantly better (at `p<alpha'`) than the second
    algorithm, then the cell is marked with `<`. If the first algorithm is
    significantly worse (at `p<alpha'`) than the second algorithm, then the
    cell is marked with `>`. If the observed differences are not significant
    (`p>=alpha'`), then the cell is marked with `?`.

    However, there could also be a situation where a statistical comparison
    makes no sense as no difference could reliably be detected anyway. For
    example, if one algorithm has a smaller median result but a larger mean
    result, or if the medians are the same, or if the means are the same.
    Regardless of what outcome a test would have, we could not really claim
    that any of the algorithms was better or worse. In such cases, no test
    is performed and `-` is printed instead (signified by `&mdash;` in the
    markdown format).

    Finally, the bottom row sums up the numbers of `<`, `?`, and `>` outcomes
    for each algorithm pair.

    Depending on the parameter `text_format_driver`, the tables can be
    rendered in different formats, such as
    :py:class:`~moptipy.utils.markdown.Markdown`,
    :py:class:`~moptipy.utils.latex.LaTeX`, and
    :py:class:`~moptipy.utils.html.HTML`.

    :param end_results: the end results data
    :param file_name: the base file name
    :param dir_name: the base directory
    :param alpha: the threshold at which the two-tailed test result is
        accepted.
    :param text_format_driver: the text format driver
    :param algorithm_sort_key: a function returning sort keys for algorithms
    :param instance_sort_key: a function returning sort keys for instances
    :param instance_namer: the name function for instances receives an
        instance ID and returns an instance name; default=identity function
    :param algorithm_namer: the name function for algorithms receives an
        algorithm ID and returns an instance name; default=identity function
    :param use_lang: should we use the language to define the filename
    :param p_renderer: the renderer for all probabilities
    :param value_getter: the getter for the values that should be compared. By
        default, the best obtained objective values are compared. However, if
        you let the runs continue until they reach a certain goal quality,
        then you may want to compare the runtimes consumed until that quality
        is reached. Basically, you can use any of the getters provided by
        :meth:`moptipy.evaluation.end_results.EndResult.getter`, but you must
        take care that the comparison makes sense, i.e., compare qualities
        under fixed-budget scenarios (the default behavior) or compare
        runtimes under scenarios with goal qualities - but do not mix up the
        scenarios.
    :returns: the path to the file with the tabulated test results
    """
    # Before doing anything, let's do some type checking on the parameters.
    # I want to ensure that this function is called correctly before we begin
    # to actually process the data. It is better to fail early than to deliver
    # some incorrect results.
    if not isinstance(end_results, Iterable):
        raise type_error(end_results, "end_results", Iterable)
    if not isinstance(file_name, str):
        raise type_error(file_name, "file_name", str)
    if not isinstance(dir_name, str):
        raise type_error(dir_name, "dir_name", str)
    if not isinstance(alpha, float):
        raise type_error(alpha, "alpha", float)
    if not (0.0 < alpha < 0.5):
        raise ValueError(f"invalid alpha={alpha}, must be 0<alpha<0.5.")
    if callable(text_format_driver):
        text_format_driver = text_format_driver()
    if not isinstance(text_format_driver, TextFormatDriver):
        raise type_error(text_format_driver, "text_format_driver",
                         TextFormatDriver, True)
    if not callable(algorithm_namer):
        raise type_error(algorithm_namer, "algorithm_namer", call=True)
    if not callable(instance_namer):
        raise type_error(instance_namer, "instance_namer", call=True)
    if not isinstance(use_lang, bool):
        raise type_error(use_lang, "use_lang", bool)

    if not isinstance(p_renderer, NumberRenderer):
        raise type_error(p_renderer, "p_renderer", NumberRenderer)
    if not callable(value_getter):
        raise type_error(value_getter, "value_getter", call=True)

    # now gather the data: algorithms -> instances -> bestF
    data_dict: dict[str, dict[str, list[int | float]]] = {}
    inst_set: set[str] = set()
    for i, end_result in enumerate(end_results):
        if not isinstance(end_result, EndResult):
            raise type_error(end_result, f"end_results[{i}]", EndResult)
        value = value_getter(end_result)
        if not isinstance(value, (int, float)):
            raise type_error(
                value, f"value_getter({end_result}", (int, float))
        if not isfinite(value):
            raise ValueError(f"value_getter({end_result}={value}")
        inst_set.add(end_result.instance)
        if end_result.algorithm in data_dict:
            inst_res = data_dict[end_result.algorithm]
        else:
            data_dict[end_result.algorithm] = inst_res = {}
        if end_result.instance in inst_res:
            inst_res[end_result.instance].append(value)
        else:
            inst_res[end_result.instance] = [value]

    # finished collecting data
    n_algos: Final[int] = len(data_dict)
    if n_algos <= 0:
        raise ValueError("end results are empty!")
    if not (1 < n_algos < 5):
        raise ValueError(f"invalid number {n_algos} of "
                         "algorithms, must be 1<n_algos<5.")
    instances: Final[list] = sorted(inst_set, key=instance_sort_key)
    n_insts: Final[int] = len(instances)
    algorithms: Final[list] = sorted(data_dict.keys(), key=algorithm_sort_key)

    # validate data
    for algo in algorithms:
        if len(data_dict[algo]) != n_insts:
            raise ValueError(
                f"algorithm '{algo}' has only data for {len(data_dict[algo])}"
                f" instances, but should have for {n_insts}.")

    # compute the bonferroni correction
    n_cols: Final[int] = ((n_algos - 1) * n_algos) // 2
    n_pairs: Final[int] = n_insts * n_cols
    alpha_prime: Final[float] = alpha / n_pairs
    if not (0.0 < alpha_prime < 0.5):
        raise ValueError(f"invalid alpha'={alpha_prime}, must be 0<alpha'"
                         f"<0.5, stems from alpha={alpha} and N={n_pairs}")

    # compute the comparison results
    comparisons: Final[list[list[int]]] = \
        [[__compare(data_dict[algorithms[i]][inst],
                    data_dict[algorithms[j]][inst])
          for inst in instances]
         for i in range(n_algos - 1)
         for j in range(i + 1, n_algos)]

    # compute the p_values
    p_values: Final[list[list[float]]] = \
        [[float(mannwhitneyu(data_dict[algorithms[ij[0]]][inst],
                             data_dict[algorithms[ij[1]]][inst]).pvalue)
          if comparisons[z][k] != 0 else None
          for k, inst in enumerate(instances)]
         for z, ij in enumerate([(ii, jj) for ii in range(n_algos - 1)
                                 for jj in range(ii + 1, n_algos)])]

    # now format all p values to strings
    p_flat: Final[list[float]] = [y for x in p_values for y in x]
    p_flat.append(alpha)
    p_flat.append(alpha_prime)
    p_flat_strs: Final[list[FormattedStr | None]] = \
        p_renderer.render(p_flat)
    alpha_str: FormattedStr = cast(FormattedStr, p_flat_strs[-2])
    a2 = FormattedStr.number(alpha)
    if len(a2) <= len(alpha_str):
        alpha_str = a2
    alpha_prime_str: FormattedStr = cast(FormattedStr, p_flat_strs[-1])
    a2 = FormattedStr.number(alpha_prime)
    if len(a2) <= len(alpha_prime_str):
        alpha_prime_str = a2
    del p_flat_strs[-1]
    del p_flat_strs[-1]
    p_strs: Final[list[list[FormattedStr | None]]] = \
        [p_flat_strs[(i * n_insts):((i + 1) * n_insts)] for i in range(n_cols)]

    summary_row: list[Iterable[str] | str] | None = None
    if n_insts > 0:
        summary_row = [__WEL]
        for i, col in enumerate(comparisons):
            wins: int = 0
            equals: int = 0
            losses: int = 0
            pv: list[float] = p_values[i]
            for j, cv in enumerate(col):
                if (cv == 0) or (pv[j] >= alpha_prime):
                    equals += 1
                elif cv < 0:
                    wins += 1
                else:
                    losses += 1
            summary_row.append(f"{wins}/{equals}/{losses}")

    #: the columns
    cols: Final[list[list[str | list[str]]]] = \
        [[__compare_to_str(comparisons[i][j], p_strs[i][j], p_values[i][j],
                           alpha_prime)
          for j in range(n_insts)]
         for i in range(n_cols)]
    cols.insert(0, [FormattedStr.add_format(instance_namer(inst), code=True)
                    for inst in instances])

    # We now got the p_values, the comparison results, and the p_strs.

    # We now can render the headlines. If there is only one pair of
    # algorithms, we render a single headline with all the information.
    # If there are multiple algorithm pairs, we do two headlines.
    head_lines: list[list[str | list[str]]]
    multi_header: Final[bool] = \
        (n_algos > 2) and (not isinstance(text_format_driver, Markdown))
    if multi_header:
        head_lines = [[], []]

        def __app(a, b) -> None:
            nonlocal head_lines
            head_lines[0].append(a)
            head_lines[1].append(b)
    else:
        head_lines = [[]]

        def __app(a, b) -> None:
            nonlocal head_lines
            x = []
            if isinstance(a, str):
                x.append(a)
            else:
                x.extend(a)
            x.append(" ")
            if isinstance(b, str):
                x.append(b)
            else:
                x.extend(b)
            head_lines[0].append(x)

    __app("Mann-Whitney U", [__ALPHA, "=", alpha_str, ", ", __ALPHA,
                             "'=", alpha_prime_str])
    for i in range(n_algos - 1):
        for j in range(i + 1, n_algos):
            an1 = FormattedStr.add_format(algorithm_namer(algorithms[i]),
                                          code=True)
            an2 = FormattedStr.add_format(algorithm_namer(algorithms[j]),
                                          code=True)
            if len(an1) >= len(an2):
                __app(an1, ["vs. ", an2])
            else:
                __app([an1, " vs."], an2)

    # write the table
    dest: Final[Path] = text_format_driver.filename(
        file_name, dir_name, use_lang)
    with dest.open_for_write() as wd:
        with Table(wd,
                   ("l" if multi_header else "r")
                   + ("c" * n_cols), text_format_driver) as table:
            with table.header() as header:
                for hl in head_lines:
                    with header.row() as row:
                        for cell in hl:
                            row.cell(cell)
            with table.section() as section:
                section.cols(cols)
            if summary_row is not None:
                with table.section() as section:
                    with section.row() as row:
                        for s in summary_row:
                            row.cell(s)

    return dest



"""
Tournament selection with or without replacement in the tournaments.

For each slot in the destination, a tournament with
:attr:`~moptipy.algorithms.modules.selections.tournament.Tournament.s`
randomly chosen participating solutions is conducted. The solution with
the best fitness wins and is copied to the destination. If
:attr:`~moptipy.algorithms.modules.selections.tournament.Tournament\
.replacement` is `True`, then the solutions are drawn with replacement
for each tournament, meaning that one solution may enter a given tournament
several times. If replacements are turned off, each solution may enter each
tournament only at most once. Either way, a solution may be selected multiple
times.

1. Peter J. B. Hancock. An Empirical Comparison of Selection Methods in
   Evolutionary Algorithms. In Terence Claus Fogarty, editor, *Selected Papers
   from the AISB Workshop on Evolutionary Computing (AISB EC'94),* April
   11-13, 1994, Leeds, UK, volume 865 of Lecture Notes in Computer Science,
   pages 80-94, Berlin/Heidelberg, Germany: Springer, ISBN: 978-3-540-58483-4.
   https://dx.doi.org/10.1007/3-540-58483-8_7. Conference organized by the
   Society for the Study of Artificial Intelligence and Simulation of
   Behaviour (AISB).
2. Uday Kumar Chakraborty and Kalyanmoy Deb and Mandira Chakraborty. Analysis
   of Selection Algorithms: A Markov Chain Approach. *Evolutionary
   Computation,* 4(2):133-167. Summer 1996. Cambridge, MA, USA: MIT Press.
   doi:10.1162/evco.1996.4.2.133.
   https://dl.acm.org/doi/pdf/10.1162/evco.1996.4.2.133
3. Tobias Blickle and Lothar Thiele. A Comparison of Selection Schemes used in
   Genetic Algorithms. Second edition, December 1995. TIK-Report 11 from the
   Eidgenössische Technische Hochschule (ETH) Zürich, Department of Electrical
   Engineering, Computer Engineering and Networks Laboratory (TIK), Zürich,
   Switzerland. ftp://ftp.tik.ee.ethz.ch/pub/publications/TIK-Report11.ps
"""

from math import inf
from typing import Any, Callable, Final, Iterable, cast

from numpy.random import Generator

from moptipy.algorithms.modules.selection import FitnessRecord, Selection
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error


# start book
class Tournament(Selection):
    """Tournament selection with or without replacement in the tournament."""

# end book
    def __init__(self, s: int = 2, replacement: bool = True) -> None:
        """
        Create the tournament selection method.

        :param s: the size of the tournaments
        :param replacement: will the tournaments be with replacement?
        """
        super().__init__()
        if not isinstance(s, int):
            raise type_error(s, "s", int)
        if s < 1:
            raise ValueError(f"Tournament size must be > 1, but is {s}.")
        if not isinstance(replacement, bool):
            raise type_error(replacement, "replacement", bool)
        #: the tournament size
        self.s: Final[int] = s
        #: should we perform replacements in the tournaments?
        self.replacement: Final[bool] = replacement

# start book
    def select(self, source: list[FitnessRecord],
               dest: Callable[[FitnessRecord], Any],
               n: int, random: Generator) -> None:
        """
        Perform deterministic best selection without replacement.

        :param source: the list with the records to select from
        :param dest: the destination collector to invoke for each selected
            record
        :param n: the number of records to select
        :param random: the random number generator
        """
        # set up tournament size=s, source length=m, replacement
        # such initialization is omitted for the sake of brevity
        # end book
        s: Final[int] = self.s
        replacement: Final[bool] = self.replacement
        m: Final[int] = len(source)

        # fast call
        choice: Final[Callable[[int, int, bool], Iterable[int]]] = \
            cast(Callable[[int, int, bool], Iterable[int]], random.choice)
        # start book
        for _ in range(n):  # conduct n tournaments
            best: FitnessRecord | None = None  # best competitor
            best_fitness: int | float = inf  # best fitness
            for i in choice(m, s, replacement):  # perform tournament
                rec = source[i]  # get record from source
                rec_fitness = rec.fitness  # get its fitness
                if rec_fitness <= best_fitness:  # if better or equal...
                    best = rec  # ... rec becomes the new best record
                    best_fitness = rec_fitness  # and remember fitness
            dest(best)  # at end of the tournament, store best in dest
        # end book

    def __str__(self):
        """
        Get the name of the tournament selection algorithm.

        :return: the name of the tournament selection algorithm
        """
        st = f"tour{self.s}"
        if self.replacement:
            return f"{st}r"
        return st

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("size", self.s)
        logger.key_value("withReplacement", self.replacement)



"""
An objective function counting the number of zeros in a bit string.

This problem exists mainly for testing purposes as counterpart of
:class:`~moptipy.examples.bitstrings.onemax.OneMax`.
"""
from typing import Final

import numba  # type: ignore
import numpy as np

from moptipy.api.objective import Objective
from moptipy.utils.types import type_error


@numba.njit(nogil=True, cache=True)
def zeromax(x: np.ndarray) -> int:
    """
    Get the length of a string minus the number of zeros in it.

    :param x: the np array
    :return: the length of the string minus the number of zeros, i.e., the
        number of ones

    >>> print(zeromax(np.array([True, True, False, False, False])))
    2
    >>> print(zeromax(np.array([True, False, True, False, False])))
    2
    >>> print(zeromax(np.array([False, True,  True, False, False])))
    2
    >>> print(zeromax(np.array([True, True, True, True, True])))
    5
    >>> print(zeromax(np.array([False, True, True, True, True])))
    4
    >>> print(zeromax(np.array([False, False, False, False, False])))
    0
    """
    return int(x.sum())


class ZeroMax(Objective):
    """Maximize the number of zeros in a bit string."""

    def __init__(self, n: int) -> None:
        """
        Initialize the zeromax objective function.

        :param n: the dimension of the problem

        >>> print(ZeroMax(2).n)
        2
        >>> print(ZeroMax(4).evaluate(np.array([True, True, False, True])))
        3
        """
        super().__init__()
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        #: the upper bound = the length of the bit strings
        self.n: Final[int] = n
        self.evaluate = zeromax  # type: ignore

    def lower_bound(self) -> int:
        """
        Get the lower bound of the zeromax objective function.

        :return: 0

        >>> print(ZeroMax(10).lower_bound())
        0
        """
        return 0

    def upper_bound(self) -> int:
        """
        Get the upper bound of the zeromax objective function.

        :return: the length of the bit string

        >>> print(ZeroMax(7).upper_bound())
        7
        """
        return self.n

    def is_always_integer(self) -> bool:
        """
        Return `True` because :func:`zeromax` always returns `int` values.

        :retval True: always
        """
        return True

    def __str__(self) -> str:
        """
        Get the name of the zeromax objective function.

        :return: `zeromax_` + length of string

        >>> print(ZeroMax(13))
        zeromax_13
        """
        return f"zeromax_{self.n}"



"""
A fully configurable, general (mu+lambda) Evolutionary Algorithm.

This evolutionary algorithm begins by sampling
:attr:`~moptipy.algorithms.so.ea.EA.mu`
solutions using the nullary search operation
:attr:`~moptipy.api.algorithm.Algorithm0.op0`. In each iteration, it then uses
:attr:`~moptipy.algorithms.so.ea.EA.mu` existing solutions as input for
the search operations, where, for each solution to be sampled, the binary
operation :attr:`~moptipy.api.algorithm.Algorithm2.op2` is used with
probability :attr:`~moptipy.algorithms.so.ea.EA.br` and (otherwise), the unary
operator :attr:`~moptipy.api.algorithm.Algorithm1` is used. The inputs of both
operators are chosen from the :attr:`~moptipy.algorithms.so.ea.EA.mu`
solutions using :attr:`~moptipy.algorithms.so.general_ea.GeneralEA.mating`
selection. After :attr:`~moptipy.algorithms.so.ea.EA.lambda_` new solutions
have been created this way (and have been evaluated as well), a fitness
assignment process (:class:`~moptipy.algorithms.so.fitness.Fitness`) assigns
fitness values to them based on their objective values
(:attr:`~moptipy.algorithms.so.record.Record.f`), maybe also using the index
of the iteration (:attr:`~moptipy.algorithms.so.record.Record.it`) in which
they were created. The survival selection
:attr:`~moptipy.algorithms.so.general_ea.GeneralEA.survival` then chooses,
from the joint set of `mu+lambda` solutions, the `mu` solutions for the
next iteration. Both mating and survival selection are instances of class
:class:`~moptipy.algorithms.modules.selection.Selection`.

This algorithm is equivalent to :class:`~moptipy.algorithms.so.ea.EA`, but
allows for using a customized fitness assignment step
(:class:`~moptipy.algorithms.so.fitness.Fitness`) as well as customizable
survival and :attr:`~moptipy.algorithms.so.general_ea.GeneralEA.mating`
selection (:class:`~moptipy.algorithms.modules.selection.Selection`).

1. Thomas Bäck, David B. Fogel, and Zbigniew Michalewicz, eds., *Handbook of
   Evolutionary Computation.* 1997. Computational Intelligence Library.
   New York, NY, USA: Oxford University Press, Inc. ISBN: 0-7503-0392-1
2. James C. Spall. *Introduction to Stochastic Search and Optimization.*
   Estimation, Simulation, and Control - Wiley-Interscience Series in Discrete
   Mathematics and Optimization, volume 6. 2003. Chichester, West Sussex, UK:
   Wiley Interscience. ISBN: 0-471-33052-3. http://www.jhuapl.edu/ISSO/.
3. Frank Hoffmeister and Thomas Bäck. Genetic Algorithms and Evolution
   Strategies: Similarities and Differences. In Hans-Paul Schwefel and
   Reinhard Männer, *Proceedings of the International Conference on Parallel
   Problem Solving from Nature (PPSN I),* October 1-3, 1990, Dortmund,
   Germany, volume 496 of Lecture Notes in Computer Science, pages 455-469,
   Berlin/Heidelberg, Germany: Springer. ISBN: 978-3-540-54148-6.
   https://doi.org/10.1007/BFb0029787.
"""
from typing import Callable, Final, cast

from numpy.random import Generator

from moptipy.algorithms.modules.selection import (
    FitnessRecord,
    Selection,
    check_selection,
)
from moptipy.algorithms.modules.selections.best import Best
from moptipy.algorithms.modules.selections.random_without_replacement import (
    RandomWithoutReplacement,
)
from moptipy.algorithms.so.ea import EA, _float_0
from moptipy.algorithms.so.fitness import Fitness, FRecord, check_fitness
from moptipy.algorithms.so.fitnesses.rank_and_iteration import RankAndIteration
from moptipy.api.operators import Op0, Op1, Op2
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import PART_SEPARATOR


class _Record(FRecord):
    """Same as :class:`FRecord`, but with a secret selection marker."""

    def __init__(self, x, f: int | float, selected: bool = False):
        """
        Create the record.

        :param x: the data structure for a point in the search space
        :param f: the corresponding objective value
        :param selected: is the record currently in use?
        """
        super().__init__(x, f)
        #: an internal flag - do NOT access!!
        self._selected: bool = selected


# start book
class GeneralEA(EA):
    """The fully customizable (mu+lambda) EA."""

    def solve(self, process: Process) -> None:
        """
        Apply the (mu+lambda) EA to an optimization problem.

        :param process: the black-box process object
        """
        # the initialization of variables is omitted for brevity
# end book
        mu: Final[int] = self.mu  # mu: number of best solutions kept
        lambda_: Final[int] = self.lambda_
        mu_plus_lambda: Final[int] = mu + lambda_  # size = mu + lambda
        random: Final[Generator] = process.get_random()  # random gen
        create: Final[Callable] = process.create  # create x container
        evaluate: Final[Callable] = process.evaluate  # the objective
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator
        op2: Final[Callable] = self.op2.op2  # the binary operator
        br: Final[float] = self.br  # the rate at which to use op2
        should_terminate: Final[Callable] = process.should_terminate
        r01: Final[Callable[[], float]] = cast(  # only if 0<br<1, we
            Callable[[], float],  # need random floats
            random.random if 0 < br < 1 else _float_0)
        assign_fitness: Final[Callable[[list[FRecord], Generator], None]] = \
            self.fitness.assign_fitness
        survival_selection: Final[Callable[
            [list[FRecord], Callable, int, Generator], None]] = \
            cast(Callable[[list[FRecord], Callable, int, Generator],
                          None], self.survival.select)
        mating_selection: Final[Callable[
            [list[FRecord], Callable, int, Generator], None]] = \
            cast(Callable[[list[FRecord], Callable, int, Generator],
                          None], self.mating.select)

        recs: Final[list] = [None] * mu_plus_lambda  # pre-allocate list
        parents: Final[list] = [None, None]  # mating pool: length 2
        population: Final[list] = [None] * mu_plus_lambda  # whole
        # Fast calls
        parents_clear: Final[Callable[[], None]] = parents.clear
        parents_append: Final[Callable[[FitnessRecord], None]] = \
            cast(Callable[[FitnessRecord], None], parents.append)
        population_clear: Final[Callable[[], None]] = population.clear
        population_append: Final[Callable[[_Record], None]] = \
            cast(Callable[[_Record], None], population.append)
# start book
        # create list of mu random records and lambda empty records
        f: int | float = 0  # variable to hold objective values
        for i in range(mu_plus_lambda):  # fill list of size mu+lambda
            x = create()  # by creating point in search space
            selected: bool = i < mu  # only fully create first mu recs
            if selected:  # only the first mu records are initialized by
                op0(random, x)  # applying nullary operator = randomize
                if should_terminate():  # should we quit?  # -book
                    return  # computational budget exhausted -> quit  # -book
                f = evaluate(x)  # continue? ok, evaluate new solution
            recs[i] = _Record(x, f, selected)  # create and store record

        mating_pool: Final[list] = recs[0:mu]  # the selection survivers
        assign_fitness(mating_pool, random)  # assign fitness first time
# end book
        mating_pool_clear: Final[Callable[[], None]] = mating_pool.clear
        mating_pool_append: Final[Callable[[FitnessRecord], None]] = \
            cast(Callable[[FitnessRecord], None], mating_pool.append)
# start book
        it: int = 0  # set the iteration counter
        while True:  # lst: keep 0..mu-1, overwrite mu..mu+lambda-1
            it = it + 1  # step the iteration counter
            population_clear()  # clear population

            di = 0  # set index of next potential destination
            for _ in range(lambda_):  # for all lambda offspring
                if should_terminate():  # only continue if we still... # -book
                    return  # ...have sufficient budget # -book
                while True:  # get the next non-selected record
                    dest = recs[di]  # get the record
                    di = di + 1  # step counter
                    if dest._selected:  # if it was selected
                        dest._selected = False  # mark it as unselected
                        population_append(dest)  # store in population
                        continue  # try next record
                    break  # use the (unselected) record as destination

                x = dest.x  # the destination "x" value
                dest.it = it  # remember iteration of solution creation
                do_binary: bool = r01() < br  # will we do binary operation?
                parents_clear()  # clear mating pool: room for 2
                mating_selection(mating_pool, parents_append,
                                 2 if do_binary else 1, random)

                if do_binary:  # binary operation (with p == br)
                    op2(random, x, parents[0].x, parents[1].x)
                else:  # unary operation otherwise
                    op1(random, x, parents[0].x)  # apply unary op
                dest.f = evaluate(x)  # evaluate new point
                population_append(dest)  # store in population

            # add remaining selected solutions from recs to population
            # from index di to mu+lambda ... omitted for brevity in book
            # end book
            for di2 in range(di, mu_plus_lambda):
                other = recs[di2]
                if other._selected:  # only if solution was selected
                    other._selected = False  # set as unselected
                    population_append(other)  # put into population
            # start book
            assign_fitness(population, random)  # assign fitness
            mating_pool_clear()  # clear list of survived records
            survival_selection(population, mating_pool_append, mu, random)
            for rec in mating_pool:  # mark all selected solutions as
                rec._selected = True  # selected
# end book

    def __init__(self, op0: Op0,
                 op1: Op1 | None = None,
                 op2: Op2 | None = None,
                 mu: int = 1, lambda_: int = 1,
                 br: float | None = None,
                 fitness: Fitness | None = None,
                 survival: Selection | None = None,
                 mating: Selection | None = None,
                 name: str = "generalEa") -> None:
        """
        Create the customizable Evolutionary Algorithm (EA).

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param op2: the binary search operator
        :param mu: the number of best solutions to survive in each generation
        :param lambda_: the number of offspring in each generation
        :param br: the rate at which the binary operator is applied
        :param fitness: the fitness assignment process
        :param survival: the survival selections algorithm
        :param mating: the mating selections algorithm
        :param name: the base name of the algorithm
        """
        if fitness is None:
            fitness = RankAndIteration()
        if fitness.__class__ is not RankAndIteration:
            name = f"{name}{PART_SEPARATOR}{fitness}"
        if survival is None:
            survival = Best()
        if mating is None:
            mating = RandomWithoutReplacement()
        if (survival.__class__ is not Best) \
                or (mating.__class__ is not RandomWithoutReplacement):
            name = f"{name}{PART_SEPARATOR}{survival}{PART_SEPARATOR}{mating}"

        super().__init__(op0, op1, op2, mu, lambda_, br, name)
        #: the fitness assignment process
        self.fitness: Final[Fitness] = check_fitness(fitness)
        #: the survival selection algorithm
        self.survival: Final[Selection] = check_selection(survival)
        #: the mating selection algorithm
        self.mating: Final[Selection] = check_selection(mating)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope("fitness") as v:
            self.fitness.log_parameters_to(v)
        with logger.scope("survival") as s:
            self.survival.log_parameters_to(s)
        with logger.scope("mating") as m:
            self.mating.log_parameters_to(m)
